/*
*  Copyright (c) 2011-2016 Samsung Electronics Co., Ltd All Rights Reserved
*  Licensed under the Apache License, Version 2.0 (the "License");
*  you may not use this file except in compliance with the License.
*  You may obtain a copy of the License at
*  http://www.apache.org/licenses/LICENSE-2.0
*  Unless required by applicable law or agreed to in writing, software
*  distributed under the License is distributed on an "AS IS" BASIS,
*  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
*  See the License for the specific language governing permissions and
*  limitations under the License.
*/

#include <app_manager.h>
#include <dirent.h>
#include <Ecore.h>
#include <iconv.h>
#include <sys/stat.h>
#include <sys/types.h>
#include <sys/wait.h>
#include <system_info.h>
#include <vconf.h>

#include "tts.h"
#include "tts_client.h"
#include "tts_config_mgr.h"
#include "tts_dbus.h"
#include "tts_main.h"

#include "tts_internal.h"

static bool g_screen_reader;

static int g_feature_enabled = -1;

static bool g_err_callback_status = false;

static int g_max_text_size = -1;

/* for repetition */
static char* g_language = NULL;

static int g_voice_type = -1;

static int g_speed = -1;


/* Function definition */
static Eina_Bool __tts_notify_state_changed(void *data);
static Eina_Bool __tts_notify_error(void *data);

const char* tts_tag()
{
	return "ttsc";
}

static int __tts_get_feature_enabled()
{
	if (0 == g_feature_enabled) {
		SLOG(LOG_ERROR, TAG_TTSC, "[ERROR] TTS feature NOT supported");
		return TTS_ERROR_NOT_SUPPORTED;
	} else if (-1 == g_feature_enabled) {
		bool tts_supported = false;
		if (0 == system_info_get_platform_bool(TTS_FEATURE_PATH, &tts_supported)) {
			if (false == tts_supported) {
				SLOG(LOG_ERROR, TAG_TTSC, "[ERROR] TTS feature NOT supported");
				g_feature_enabled = 0;
				return TTS_ERROR_NOT_SUPPORTED;
			}

			g_feature_enabled = 1;
		} else {
			SLOG(LOG_ERROR, TAG_TTSC, "[ERROR] Fail to get feature value");
			return TTS_ERROR_NOT_SUPPORTED;
		}
	}

	return 0;
}

static const char* __tts_get_error_code(tts_error_e err)
{
	switch (err) {
	case TTS_ERROR_NONE:			return "TTS_ERROR_NONE";
	case TTS_ERROR_OUT_OF_MEMORY:		return "TTS_ERROR_OUT_OF_MEMORY";
	case TTS_ERROR_IO_ERROR:		return "TTS_ERROR_IO_ERROR";
	case TTS_ERROR_INVALID_PARAMETER:	return "TTS_ERROR_INVALID_PARAMETER";
	case TTS_ERROR_OUT_OF_NETWORK:		return "TTS_ERROR_OUT_OF_NETWORK";
	case TTS_ERROR_TIMED_OUT:		return "TTS_ERROR_TIMED_OUT";
	case TTS_ERROR_PERMISSION_DENIED:	return "TTS_ERROR_PERMISSION_DENIED";
	case TTS_ERROR_NOT_SUPPORTED:		return "TTS_ERROR_NOT_SUPPORTED";
	case TTS_ERROR_INVALID_STATE:		return "TTS_ERROR_INVALID_STATE";
	case TTS_ERROR_INVALID_VOICE:		return "TTS_ERROR_INVALID_VOICE";
	case TTS_ERROR_ENGINE_NOT_FOUND:	return "TTS_ERROR_ENGINE_NOT_FOUND";
	case TTS_ERROR_OPERATION_FAILED:	return "TTS_ERROR_OPERATION_FAILED";
	case TTS_ERROR_AUDIO_POLICY_BLOCKED:	return "TTS_ERROR_AUDIO_POLICY_BLOCKED";
	case TTS_ERROR_NOT_SUPPORTED_FEATURE:	return "TTS_ERROR_NOT_SUPPORTED_FEATURE";
	case TTS_ERROR_SERVICE_RESET:		return "TTS_ERROR_SERVICE_RESET";
	default:
		return "Invalid error code";
	}
	return NULL;
}

static int __tts_convert_config_error_code(tts_config_error_e code)
{
	if (code == TTS_CONFIG_ERROR_NONE)			return TTS_ERROR_NONE;
	if (code == TTS_CONFIG_ERROR_OUT_OF_MEMORY)		return TTS_ERROR_OUT_OF_MEMORY;
	if (code == TTS_CONFIG_ERROR_IO_ERROR)			return TTS_ERROR_IO_ERROR;
	if (code == TTS_CONFIG_ERROR_INVALID_PARAMETER)		return TTS_ERROR_INVALID_PARAMETER;
	if (code == TTS_CONFIG_ERROR_INVALID_STATE)		return TTS_ERROR_INVALID_STATE;
	if (code == TTS_CONFIG_ERROR_INVALID_VOICE)		return TTS_ERROR_INVALID_VOICE;
	if (code == TTS_CONFIG_ERROR_ENGINE_NOT_FOUND)		return TTS_ERROR_ENGINE_NOT_FOUND;
	if (code == TTS_CONFIG_ERROR_OPERATION_FAILED)		return TTS_ERROR_OPERATION_FAILED;
	if (code == TTS_CONFIG_ERROR_NOT_SUPPORTED_FEATURE)	return TTS_ERROR_NOT_SUPPORTED_FEATURE;

	return code;
}

void __tts_config_voice_changed_cb(const char* before_lang, int before_voice_type, const char* language, int voice_type, bool auto_voice, void* user_data)
{
	SLOG(LOG_DEBUG, TAG_TTSC, "Voice changed : Before lang(%s) type(%d) , Current lang(%s), type(%d)",
		before_lang, before_voice_type, language, voice_type);

	GList* client_list = NULL;
	client_list = tts_client_get_client_list();

	GList *iter = NULL;
	tts_client_s *data = NULL;

	if (g_list_length(client_list) > 0) {
		/* Get a first item */
		iter = g_list_first(client_list);

		while (NULL != iter) {
			data = iter->data;
			if (NULL != data->default_voice_changed_cb) {
				SLOG(LOG_DEBUG, TAG_TTSC, "Call default voice changed callback : uid(%d)", data->uid);
				data->default_voice_changed_cb(data->tts, before_lang, before_voice_type, 
					language, voice_type, data->default_voice_changed_user_data);
			}

			/* Check whether language is changed or not. If it is changed, make 'text_repeat' NULL */
			if (0 != strncmp(before_lang, language, strlen(before_lang))) {
				if (NULL != data->text_repeat) {
					free(data->text_repeat);
					data->text_repeat = NULL;
				}
			}

			/* Next item */
			iter = g_list_next(iter);
		}
	}

	return;
}

static Eina_Bool __reconnect_by_engine_changed(void* data)
{
	tts_h tts = (tts_h)data;

	tts_client_s* client = tts_client_get(tts);
	if (NULL == client) {
		SLOG(LOG_ERROR, TAG_TTSC, "[WARNING] A handle is not valid");
		return EINA_FALSE;
	}

	if (TTS_STATE_READY != client->current_state) {
		usleep(10000);
		return EINA_TRUE;
	}

	int ret = tts_unprepare(tts);
	if (0 != ret) {
		SLOG(LOG_ERROR, TAG_TTSC, "[ERROR] Fail to prepare for setting a new engine... (%d)", ret);
	}
	ret = tts_prepare(tts);
	if (0 != ret) {
		SLOG(LOG_ERROR, TAG_TTSC, "[ERROR] Fail to prepare for setting a new engine... (%d)", ret);
	}

	return EINA_FALSE;
}

void _tts_config_engine_changed_cb(const char* engine_id, const char* setting, const char* language, int voice_type, bool auto_voice, bool need_credential, void* user_data)
{
	tts_h tts = (tts_h)user_data;

	tts_client_s* client = tts_client_get(tts);
	if (NULL == client) {
		SLOG(LOG_ERROR, TAG_TTSC, "[WARNING] A handle is not valid");
		return;
	}

	if (NULL != engine_id)	SLOG(LOG_DEBUG, TAG_TTSC, "Engine id(%s)", engine_id);
	if (NULL != setting)	SLOG(LOG_DEBUG, TAG_TTSC, "Engine setting(%s)", setting);
	if (NULL != language)	SLOG(LOG_DEBUG, TAG_TTSC, "Language(%s)", language);
	SLOG(LOG_DEBUG, TAG_TTSC, "Voice type(%d), Auto voice(%s), Credential(%s)", voice_type, auto_voice ? "on" : "off", need_credential ? "need" : "no need");

	/* When the default engine is changed, please unload the old engine and load the new one. */
	int ret = -1;

	if (TTS_STATE_PLAYING == client->current_state || TTS_STATE_PAUSED == client->current_state) {
		ret = tts_stop(tts);
		if (0 != ret) {
			SLOG(LOG_DEBUG, TAG_TTSC, "[DEBUG] TTS client stopping...");
		}

		ecore_idler_add(__reconnect_by_engine_changed, (void*)tts);
	} else if (TTS_STATE_READY == client->current_state) {
		ret = tts_unprepare(tts);
		if (0 != ret) {
			SLOG(LOG_ERROR, TAG_TTSC, "[ERROR] Fail to unprepare for setting a new engine... (%d)", ret);
		}
		ret = tts_prepare(tts);
		if (0 != ret) {
			SLOG(LOG_ERROR, TAG_TTSC, "[ERROR] Fail to prepare for setting a new engine... (%d)", ret);
		}
	}

	/* call callback function */
	if (NULL != client->engine_changed_cb) {
		client->engine_changed_cb(tts, engine_id, language, voice_type, need_credential, client->engine_changed_user_data);
	} else {
		SLOG(LOG_WARN, TAG_TTSC, "No registered callback function for changed engine");
	}
	return;
}

int tts_create(tts_h* tts)
{
	if (0 != __tts_get_feature_enabled()) {
		return TTS_ERROR_NOT_SUPPORTED;
	}

	SLOG(LOG_INFO, TAG_TTSC, "@@@ Create TTS");

	/* check param */
	if (NULL == tts) {
		SLOG(LOG_ERROR, TAG_TTSC, "[ERROR] Input handle is null");
		SLOG(LOG_DEBUG, TAG_TTSC, "@@@");
		return TTS_ERROR_INVALID_PARAMETER;
	}

	if (0 == tts_client_get_size()) {
		if (0 != tts_dbus_open_connection()) {
			SLOG(LOG_ERROR, TAG_TTSC, "[ERROR] Fail to open dbus connection");
			return TTS_ERROR_OPERATION_FAILED;
		}
	}

	if (0 != tts_client_new(tts)) {
		SLOG(LOG_ERROR, TAG_TTSC, "[ERROR] Fail to create client!!!!!");
		return TTS_ERROR_OUT_OF_MEMORY;
	}

	tts_client_s* client = tts_client_get(*tts);
	if (NULL == client) {
		SLOG(LOG_ERROR, TAG_TTSC, "[ERROR] Fail to get client");
		return TTS_ERROR_OPERATION_FAILED;
	}

	int ret = tts_config_mgr_initialize(client->uid);
	if (0 != ret) {
		SLOG(LOG_ERROR, TAG_TTSC, "[ERROR] Fail to init config manager : %d", ret);
		tts_client_destroy(*tts);
		return __tts_convert_config_error_code(ret);
	}

	ret = tts_config_mgr_set_callback(client->uid, _tts_config_engine_changed_cb, __tts_config_voice_changed_cb, NULL, NULL, client->tts);
	if (0 != ret) {
		SLOG(LOG_ERROR, TAG_TTSC, "[ERROR] Fail to set config changed : %d", ret);
		tts_client_destroy(*tts);
		return __tts_convert_config_error_code(ret);
	}

	SLOG(LOG_DEBUG, TAG_TTSC, "@@@");
	SLOG(LOG_DEBUG, TAG_TTSC, " ");

	return TTS_ERROR_NONE;
}

int tts_destroy(tts_h tts)
{
	if (0 != __tts_get_feature_enabled()) {
		return TTS_ERROR_NOT_SUPPORTED;
	}

	SLOG(LOG_INFO, TAG_TTSC, "@@@ Destroy TTS");

	if (NULL == tts) {
		SLOG(LOG_ERROR, TAG_TTSC, "[ERROR] Input handle is null");
		return TTS_ERROR_INVALID_PARAMETER;
	}

	tts_client_s* client = tts_client_get(tts);

	/* check handle */
	if (NULL == client) {
		SLOG(LOG_ERROR, TAG_TTSC, "[ERROR] A handle is not valid");
		return TTS_ERROR_INVALID_PARAMETER;
	}

	/* check used callback */
	if (0 != tts_client_get_use_callback(client)) {
		SLOG(LOG_ERROR, TAG_TTSC, "[ERROR] Cannot destroy in Callback function");
		return TTS_ERROR_OPERATION_FAILED;
	}

	tts_config_mgr_finalize(client->uid);

	int ret = -1;
	int count = 0;
	int screen_reader = -1;

	/* check state */
	switch (client->current_state) {
	case TTS_STATE_PAUSED:
	case TTS_STATE_PLAYING:
	case TTS_STATE_READY:
		ret = vconf_get_bool(TTS_ACCESSIBILITY_KEY, &screen_reader);
		if (0 != ret) {
			SLOG(LOG_ERROR, tts_tag(), "[ERROR] Fail to get screen reader");
		} else {
			SLOG(LOG_INFO, tts_tag(), "[INFO] Success to get screen reader(%d), g_screen_reader(%s), client->mode(%d)", screen_reader, (true == g_screen_reader) ? "True" : "False", client->mode);
			g_screen_reader = (bool)screen_reader;
		}
		if (!(false == g_screen_reader && TTS_MODE_SCREEN_READER == client->mode)) {
			do {
				ret = tts_dbus_request_finalize(client->uid);
				if (0 != ret) {
					if (TTS_ERROR_TIMED_OUT != ret) {
						SLOG(LOG_ERROR, TAG_TTSC, "[ERROR] result : %s", __tts_get_error_code(ret));
						break;
					} else {
						SLOG(LOG_WARN, TAG_TTSC, "[WARNING] retry finalize");
						usleep(10000);
						count++;
						if (TTS_RETRY_COUNT == count) {
							SLOG(LOG_ERROR, TAG_TTSC, "[ERROR] Fail to request");
							break;
						}
					}
				}
			} while (0 != ret);
		} else {
			SLOG(LOG_WARN, TAG_TTSC, "[WARNING] Do not request finalize : g_sr(%d) mode(%d)", g_screen_reader, client->mode);
		}

		client->before_state = client->current_state;
		client->current_state = TTS_STATE_CREATED;

	case TTS_STATE_CREATED:
		if (NULL != client->conn_timer) {
			SLOG(LOG_DEBUG, TAG_TTSC, "Connect Timer is deleted");
			ecore_timer_del(client->conn_timer);
			client->conn_timer = NULL;
		}
		/* Free resources */
		tts_client_destroy(tts);
		break;

	default:
		break;
	}

	if (0 == tts_client_get_size()) {
		if (0 != tts_dbus_close_connection()) {
			SLOG(LOG_ERROR, TAG_TTSC, "[ERROR] Fail to close connection");
		}
	}

	if (NULL != g_language) {
		free(g_language);
		g_language = NULL;
	}

	tts = NULL;

	SLOG(LOG_DEBUG, TAG_TTSC, "@@@");

	return TTS_ERROR_NONE;
}

void __tts_screen_reader_changed_cb(bool value)
{
	g_screen_reader = value;
}

int tts_set_mode(tts_h tts, tts_mode_e mode)
{
	if (0 != __tts_get_feature_enabled()) {
		return TTS_ERROR_NOT_SUPPORTED;
	}

	SLOG(LOG_INFO, TAG_TTSC, "@@@ Set TTS mode(%d)", mode);

	tts_client_s* client = tts_client_get(tts);

	/* check handle */
	if (NULL == client) {
		SLOG(LOG_ERROR, TAG_TTSC, "[ERROR] A handle is not available");
		SLOG(LOG_DEBUG, TAG_TTSC, "@@@");
		return TTS_ERROR_INVALID_PARAMETER;
	}

	/* check state */
	if (client->current_state != TTS_STATE_CREATED) {
		SLOG(LOG_ERROR, TAG_TTSC, "[ERROR] Invalid State: Current state is not 'CREATED'");
		SLOG(LOG_DEBUG, TAG_TTSC, "@@@");
		SLOG(LOG_DEBUG, TAG_TTSC, " ");
		return TTS_ERROR_INVALID_STATE;
	}

	if (TTS_MODE_DEFAULT <= mode && mode <= TTS_MODE_INTERRUPT) {
		client->mode = mode;
	} else {
		SLOG(LOG_ERROR, TAG_TTSC, "[ERROR] mode is not valid : %d", mode);
		SLOG(LOG_DEBUG, TAG_TTSC, "@@@");
		SLOG(LOG_DEBUG, TAG_TTSC, " ");
		return TTS_ERROR_INVALID_PARAMETER;
	}

	if (TTS_MODE_SCREEN_READER == mode) {
		int ret;
		int screen_reader;
		ret = vconf_get_bool(TTS_ACCESSIBILITY_KEY, &screen_reader);
		if (0 != ret) {
			SLOG(LOG_ERROR, tts_tag(), "[Config ERROR] Fail to get screen reader");
			return TTS_ERROR_OPERATION_FAILED;
		}
		g_screen_reader = (bool)screen_reader;
		tts_config_set_screen_reader_callback(client->uid, __tts_screen_reader_changed_cb);
	} else {
		tts_config_unset_screen_reader_callback(client->uid);
	}

	SLOG(LOG_DEBUG, TAG_TTSC, "@@@");

	return TTS_ERROR_NONE;
}

int tts_get_mode(tts_h tts, tts_mode_e* mode)
{
	if (0 != __tts_get_feature_enabled()) {
		return TTS_ERROR_NOT_SUPPORTED;
	}

	SLOG(LOG_DEBUG, TAG_TTSC, "@@@ Get TTS mode");

	tts_client_s* client = tts_client_get(tts);

	/* check handle */
	if (NULL == client) {
		SLOG(LOG_ERROR, TAG_TTSC, "[ERROR] A handle is not available");
		SLOG(LOG_DEBUG, TAG_TTSC, "@@@");
		SLOG(LOG_DEBUG, TAG_TTSC, " ");
		return TTS_ERROR_INVALID_PARAMETER;
	}

	/* check state */
	if (client->current_state != TTS_STATE_CREATED) {
		SLOG(LOG_ERROR, TAG_TTSC, "[ERROR] Invalid State: Current state is not 'CREATED'");
		SLOG(LOG_DEBUG, TAG_TTSC, "@@@");
		SLOG(LOG_DEBUG, TAG_TTSC, " ");
		return TTS_ERROR_INVALID_STATE;
	}

	if (NULL == mode) {
		SLOG(LOG_ERROR, TAG_TTSC, "[ERROR] Input parameter(mode) is NULL");
		return TTS_ERROR_INVALID_PARAMETER;
	}

	*mode = client->mode;

	SLOG(LOG_DEBUG, TAG_TTSC, "@@@");
	SLOG(LOG_DEBUG, TAG_TTSC, " ");

	return TTS_ERROR_NONE;
}

int tts_set_credential(tts_h tts, const char* credential)
{
	if (0 != __tts_get_feature_enabled()) {
		return TTS_ERROR_NOT_SUPPORTED;
	}

	if (NULL == tts || NULL == credential) {
		SLOG(LOG_ERROR, TAG_TTSC, "[ERROR] Input parameter is null");
		return TTS_ERROR_INVALID_PARAMETER;
	}

	tts_client_s* client = tts_client_get(tts);

	if (NULL == client) {
		SLOG(LOG_ERROR, TAG_TTSC, "[ERROR] Get state : A handle is not valid");
		return TTS_ERROR_INVALID_PARAMETER;
	}

	if (TTS_STATE_CREATED != client->current_state && TTS_STATE_READY != client->current_state) {
		SLOG(LOG_ERROR, TAG_TTSC, "[ERROR] The current state is invalid (%d).", client->current_state);
		return TTS_ERROR_INVALID_STATE;
	}

	if (NULL != client->credential) {
		free(client->credential);
		client->credential = NULL;
	}
	client->credential = strdup(credential);

	SLOG(LOG_DEBUG, TAG_TTSC, "@@@");

	return TTS_ERROR_NONE;
}

int tts_set_server_tts(tts_h tts, const char* credential)
{
	if (0 != __tts_get_feature_enabled()) {
		return TTS_ERROR_NOT_SUPPORTED;
	}

	if (NULL == tts) {
		SLOG(LOG_ERROR, TAG_TTSC, "[ERROR] Input parameter is null");
		return TTS_ERROR_INVALID_PARAMETER;
	}

	tts_client_s* client = tts_client_get(tts);

	if (NULL == client) {
		SLOG(LOG_ERROR, TAG_TTSC, "[ERROR] Get state : A handle is not valid");
		return TTS_ERROR_INVALID_PARAMETER;
	}

	if (TTS_STATE_CREATED != client->current_state && TTS_STATE_READY != client->current_state) {
		SLOG(LOG_ERROR, TAG_TTSC, "[ERROR] The current state is invalid (%d).", client->current_state);
		return TTS_ERROR_INVALID_STATE;
	}

	if (NULL != client->credential) {
		free(client->credential);
		client->credential = NULL;
	}

	client->internal = true;

	char* key = NULL;
	if (NULL != credential) {
		key = strdup("EnableServerTTS");
		client->credential = strdup(credential);
		if (NULL == client->credential) {
			SLOG(LOG_ERROR, TAG_TTSC, "[ERROR] Fail to allocate memory");
			if (NULL != key) {
				free(key);
				key = NULL;
			}
			return TTS_ERROR_OUT_OF_MEMORY;
		}
	} else {
		key = strdup("DisableServerTTS");
	}

	if (NULL == key) {
		SLOG(LOG_ERROR, TAG_TTSC, "[ERROR] Fail to allocate memory");
		return TTS_ERROR_OUT_OF_MEMORY;
	}

	int pid = getpid();
	char* appid = NULL;
	int ret = app_manager_get_app_id(pid, &appid);
	if (0 != ret || NULL == appid) {
		SLOG(LOG_ERROR, TAG_TTSC, "[ERROR] Fail to get appid, ret(%d), pid(%d), appid(%s)", ret, pid, appid);
		free(key);
		key = NULL;
		if (NULL != appid) {
			free(appid);
			appid = NULL;
		}
		return TTS_ERROR_OPERATION_FAILED;
	}

	ret = tts_set_private_data(tts, key, appid);
	if (0 != ret) {
		SLOG(LOG_ERROR, TAG_TTSC, "[ERROR] Fail to set private data, ret(%d), pid(%d), appid(%s)", ret, pid, appid);
		free(key);
		key = NULL;
		free(appid);
		appid = NULL;
		return ret;
	}

	free(appid);
	appid = NULL;
	free(key);
	key = NULL;

	SLOG(LOG_DEBUG, TAG_TTSC, "@@@");

	return TTS_ERROR_NONE;
}

static Eina_Bool __tts_connect_daemon(void *data)
{
	tts_h tts = (tts_h)data;
	tts_client_s* client = tts_client_get(tts);

	/* check handle */
	if (NULL == client) {
		SLOG(LOG_ERROR, TAG_TTSC, "[ERROR] A handle is not valid");
		return EINA_FALSE;
	}

	/* Send hello */
	if (0 != tts_dbus_request_hello(client->uid)) {
		return EINA_TRUE;
	}

	SLOG(LOG_INFO, TAG_TTSC, "@@@ Connect daemon");

	/* do request initialize */
	int ret = -1;
	bool credential_needed = false;

	ret = tts_dbus_request_initialize(client->uid, &credential_needed);

	if (TTS_ERROR_ENGINE_NOT_FOUND == ret) {
		SLOG(LOG_ERROR, TAG_TTSC, "[ERROR] Fail to initialize : %s", __tts_get_error_code(ret));

		client->reason = TTS_ERROR_ENGINE_NOT_FOUND;
		client->utt_id = -1;

		ecore_timer_add(0, __tts_notify_error, (void*)client->tts);
		client->conn_timer = NULL;
		return EINA_FALSE;

	} else if (TTS_ERROR_PERMISSION_DENIED == ret) {
		SLOG(LOG_ERROR, TAG_TTSC, "[ERROR] Fail to initialize : %s", __tts_get_error_code(ret));

		client->reason = TTS_ERROR_PERMISSION_DENIED;
		client->utt_id = -1;

		ecore_timer_add(0, __tts_notify_error, (void*)client->tts);
		client->conn_timer = NULL;
		return EINA_FALSE;

	} else if (TTS_ERROR_NONE != ret) {
		SLOG(LOG_WARN, TAG_TTSC, "[WARNING] Fail to connection. Retry to connect : %s", __tts_get_error_code(ret));
		return EINA_TRUE;

	} else {
		/* success to connect tts-daemon */
		client->credential_needed = credential_needed;
		SLOG(LOG_ERROR, TAG_TTSC, "Supported options : credential(%s)", credential_needed ? "need" : "no need");
	}

	client->conn_timer = NULL;

	client = tts_client_get(tts);
	/* check handle */
	if (NULL == client) {
		SLOG(LOG_ERROR, TAG_TTSC, "[ERROR] A handle is not valid");
		return EINA_FALSE;
	}

	client->before_state = client->current_state;
	client->current_state = TTS_STATE_READY;

	if (NULL != client->state_changed_cb) {
		tts_client_use_callback(client);
		client->state_changed_cb(client->tts, client->before_state, client->current_state, client->state_changed_user_data);
		tts_client_not_use_callback(client);
	} else {
		SLOG(LOG_WARN, TAG_TTSC, "State changed callback is NULL");
	}

	SLOG(LOG_DEBUG, TAG_TTSC, "@@@");

	return EINA_FALSE;
}

int tts_prepare(tts_h tts)
{
	if (0 != __tts_get_feature_enabled()) {
		return TTS_ERROR_NOT_SUPPORTED;
	}

	SLOG(LOG_INFO, TAG_TTSC, "@@@ Prepare TTS");

	tts_client_s* client = tts_client_get(tts);

	/* check handle */
	if (NULL == client) {
		SLOG(LOG_ERROR, TAG_TTSC, "[ERROR] A handle is not available");
		SLOG(LOG_DEBUG, TAG_TTSC, "@@@");
		return TTS_ERROR_INVALID_PARAMETER;
	}

	/* check state */
	if (client->current_state != TTS_STATE_CREATED) {
		SLOG(LOG_ERROR, TAG_TTSC, "[ERROR] Invalid State: Current state is not 'CREATED'");
		SLOG(LOG_DEBUG, TAG_TTSC, "@@@");
		return TTS_ERROR_INVALID_STATE;
	}

	ecore_thread_main_loop_begin();
	client->conn_timer = ecore_timer_add(0.02, __tts_connect_daemon, (void*)tts);
	ecore_thread_main_loop_end();

	SLOG(LOG_DEBUG, TAG_TTSC, "@@@");

	return TTS_ERROR_NONE;
}

int tts_prepare_sync(tts_h tts)
{
	if (0 != __tts_get_feature_enabled()) {
		return TTS_ERROR_NOT_SUPPORTED;
	}

	SLOG(LOG_INFO, TAG_TTSC, "@@@ Prepare TTS");

	tts_client_s* client = tts_client_get(tts);

	/* check handle */
	if (NULL == client) {
		SLOG(LOG_ERROR, TAG_TTSC, "[ERROR] A handle is not available");
		SLOG(LOG_DEBUG, TAG_TTSC, "@@@");
		return TTS_ERROR_INVALID_PARAMETER;
	}

	/* check state */
	if (client->current_state != TTS_STATE_CREATED) {
		SLOG(LOG_ERROR, TAG_TTSC, "[ERROR] Invalid State: Current state is not 'CREATED'");
		SLOG(LOG_DEBUG, TAG_TTSC, "@@@");
		return TTS_ERROR_INVALID_STATE;
	}

	int cnt = 0;
	while (EINA_TRUE == __tts_connect_daemon((void*)tts) && TTS_CONNECTION_RETRY_COUNT > cnt) {
		cnt++;
	}

	SLOG(LOG_DEBUG, TAG_TTSC, "@@@");

	if (TTS_CONNECTION_RETRY_COUNT == cnt) {
		SLOG(LOG_ERROR, TAG_TTSC, "[ERROR] Fail to connect daemon");
		return TTS_ERROR_OPERATION_FAILED;
	}

	return TTS_ERROR_NONE;
}

int tts_unprepare(tts_h tts)
{
	if (0 != __tts_get_feature_enabled()) {
		return TTS_ERROR_NOT_SUPPORTED;
	}

	SLOG(LOG_INFO, TAG_TTSC, "@@@ Unprepare TTS");

	tts_client_s* client = tts_client_get(tts);

	/* check handle */
	if (NULL == client) {
		SLOG(LOG_ERROR, TAG_TTSC, "[ERROR] A handle is not available");
		return TTS_ERROR_INVALID_PARAMETER;
	}

	/* check state */
	if (client->current_state != TTS_STATE_READY) {
		SLOG(LOG_ERROR, TAG_TTSC, "[ERROR] Invalid State: Current state is not 'READY'");
		return TTS_ERROR_INVALID_STATE;
	}

	int ret = -1;
	int count = 0;
	int screen_reader = -1;

	ret = vconf_get_bool(TTS_ACCESSIBILITY_KEY, &screen_reader);
	if (0 != ret) {
		SLOG(LOG_ERROR, tts_tag(), "[ERROR] Fail to get screen reader");
	} else {
		SLOG(LOG_INFO, tts_tag(), "[INFO] Success to get screen reader(%d), g_screen_reader(%s), client->mode(%d)", screen_reader, (true == g_screen_reader) ? "True" : "False", client->mode);
		g_screen_reader = (bool)screen_reader;
	}

	bool is_prepared = false;
	if (!(false == g_screen_reader && TTS_MODE_SCREEN_READER == client->mode)) {
		do {
			ret = tts_dbus_request_finalize(client->uid);
			if (0 != ret) {
				if (TTS_ERROR_INVALID_PARAMETER == ret && false == is_prepared) {
					client->current_state = TTS_STATE_CREATED;
					if (0 == tts_prepare_sync(tts)) {
						is_prepared = true;
						SLOG(LOG_INFO, TAG_TTSC, "[INFO] Success tts_prepare_sync");
					}
				} else if (TTS_ERROR_TIMED_OUT != ret) {
					SLOG(LOG_ERROR, TAG_TTSC, "[ERROR] result : %s", __tts_get_error_code(ret));
					break;
				} else {
					SLOG(LOG_WARN, TAG_TTSC, "[WARNING] retry finalize : %s", __tts_get_error_code(ret));
					usleep(10000);
					count++;
					if (TTS_RETRY_COUNT == count) {
						SLOG(LOG_ERROR, TAG_TTSC, "[ERROR] Fail to request");
						break;
					}
				}
			}
		} while (0 != ret);
	} else {
		SLOG(LOG_WARN, TAG_TTSC, "[WARNING] Do not request finalize : g_sr(%d) mode(%d)", g_screen_reader, client->mode);
	}

	client->before_state = client->current_state;
	client->current_state = TTS_STATE_CREATED;

	if (NULL != client->state_changed_cb) {
		tts_client_use_callback(client);
		client->state_changed_cb(client->tts, client->before_state, client->current_state, client->state_changed_user_data);
		tts_client_not_use_callback(client);
		SLOG(LOG_DEBUG, TAG_TTSC, "State changed callback is called");
	}

	SLOG(LOG_DEBUG, TAG_TTSC, "@@@");

	return TTS_ERROR_NONE;
}

bool __tts_supported_voice_cb(const char* engine_id, const char* language, int type, void* user_data)
{
	tts_h tts = (tts_h)user_data;

	tts_client_s* client = tts_client_get(tts);
	if (NULL == client) {
		SLOG(LOG_ERROR, TAG_TTSC, "[WARNING] A handle is not valid");
		return false;
	}

	/* call callback function */
	if (NULL != client->supported_voice_cb) {
		return client->supported_voice_cb(tts, language, type, client->supported_voice_user_data);
	} else {
		SLOG(LOG_WARN, TAG_TTSC, "No registered callback function of supported voice");
	}

	return false;
}

int tts_foreach_supported_voices(tts_h tts, tts_supported_voice_cb callback, void* user_data)
{
	if (0 != __tts_get_feature_enabled()) {
		return TTS_ERROR_NOT_SUPPORTED;
	}

	SLOG(LOG_DEBUG, TAG_TTSC, "@@@ Foreach supported voices");

	if (NULL == tts || NULL == callback) {
		SLOG(LOG_ERROR, TAG_TTSC, "[ERROR] Input parameter is null");
		SLOG(LOG_DEBUG, TAG_TTSC, "@@@");
		return TTS_ERROR_INVALID_PARAMETER;
	}

	tts_client_s* client = tts_client_get(tts);

	/* check handle */
	if (NULL == client) {
		SLOG(LOG_ERROR, TAG_TTSC, "[ERROR] A handle is not valid");
		SLOG(LOG_DEBUG, TAG_TTSC, "@@@");
		return TTS_ERROR_INVALID_PARAMETER;
	}

	int ret = 0;
	char* current_engine = NULL;
	ret = tts_config_mgr_get_engine(&current_engine);
	if (0 != ret) {
		SLOG(LOG_ERROR, TAG_TTSC, "[ERROR] Fail to get current engine : %d", ret);
		return __tts_convert_config_error_code(ret);
	}

	client->supported_voice_cb = callback;
	client->supported_voice_user_data = user_data;

	ret = tts_config_mgr_get_voice_list(current_engine, __tts_supported_voice_cb, client->tts);

	if (NULL != current_engine) {
		free(current_engine);
		current_engine = NULL;
	}

	client->supported_voice_cb = NULL;
	client->supported_voice_user_data = NULL;

	if (0 != ret) {
		SLOG(LOG_ERROR, TAG_TTSC, "[ERROR] Result : %d", ret);
		ret = TTS_ERROR_OPERATION_FAILED;
	}

	SLOG(LOG_DEBUG, TAG_TTSC, "@@@");

	return ret;
}

int tts_get_default_voice(tts_h tts, char** lang, int* vctype)
{
	if (0 != __tts_get_feature_enabled()) {
		return TTS_ERROR_NOT_SUPPORTED;
	}

	SLOG(LOG_DEBUG, TAG_TTSC, "@@@ Get default voice");

	if (NULL == tts) {
		SLOG(LOG_ERROR, TAG_TTSC, "[ERROR] Input handle is null");
		SLOG(LOG_DEBUG, TAG_TTSC, "@@@");
		return TTS_ERROR_INVALID_PARAMETER;

	}
	tts_client_s* client = tts_client_get(tts);

	if (NULL == client) {
		SLOG(LOG_ERROR, TAG_TTSC, "[ERROR] A handle is not valid");
		SLOG(LOG_DEBUG, TAG_TTSC, "@@@");
		return TTS_ERROR_INVALID_PARAMETER;
	}

	/* Request call remote method */
	int ret = 0;
	ret = tts_config_mgr_get_voice(lang, vctype);
	if (0 != ret) {
		SLOG(LOG_ERROR, TAG_TTSC, "[ERROR] result : %d", ret);
		return __tts_convert_config_error_code(ret);
	} else {
		SLOG(LOG_DEBUG, TAG_TTSC, "[DEBUG] Default language(%s), type(%d)", *lang, *vctype);
	}

	SLOG(LOG_DEBUG, TAG_TTSC, "@@@");

	return ret;
}

int tts_get_max_text_size(tts_h tts, unsigned int* size)
{
	if (0 != __tts_get_feature_enabled()) {
		return TTS_ERROR_NOT_SUPPORTED;
	}

	if (NULL == tts || NULL == size) {
		SLOG(LOG_ERROR, TAG_TTSC, "[ERROR] Get max text count : Input parameter is null");
		return TTS_ERROR_INVALID_PARAMETER;
	}

	tts_client_s* client = tts_client_get(tts);

	if (NULL == client) {
		SLOG(LOG_ERROR, TAG_TTSC, "[ERROR] Get max text count : A handle is not valid");
		return TTS_ERROR_INVALID_PARAMETER;
	}

	if (TTS_STATE_READY != client->current_state) {
		SLOG(LOG_ERROR, TAG_TTSC, "[ERROR] Get max text count : Current state is NOT 'READY'.");
		return TTS_ERROR_INVALID_STATE;
	}

	if (0 != tts_config_mgr_get_max_text_size(size)) {
		SLOG(LOG_ERROR, TAG_TTSC, "[ERROR] Fail to get max text size");
		return TTS_ERROR_INVALID_PARAMETER;
	}

	g_max_text_size = (int)*size;

	SLOG(LOG_DEBUG, TAG_TTSC, "Get max text count : %d", *size);
	return TTS_ERROR_NONE;
}

int tts_get_state(tts_h tts, tts_state_e* state)
{
	if (0 != __tts_get_feature_enabled()) {
		return TTS_ERROR_NOT_SUPPORTED;
	}

	if (NULL == tts || NULL == state) {
		SLOG(LOG_ERROR, TAG_TTSC, "[ERROR] Get state : Input parameter is null");
		return TTS_ERROR_INVALID_PARAMETER;
	}

	tts_client_s* client = tts_client_get(tts);

	if (NULL == client) {
		SLOG(LOG_ERROR, TAG_TTSC, "[ERROR] Get state : A handle is not valid");
		return TTS_ERROR_INVALID_PARAMETER;
	}

	*state = client->current_state;

	switch (*state) {
	case TTS_STATE_CREATED:	SLOG(LOG_DEBUG, TAG_TTSC, "Current state is 'Created'");	break;
	case TTS_STATE_READY:	SLOG(LOG_DEBUG, TAG_TTSC, "Current state is 'Ready'");		break;
	case TTS_STATE_PLAYING:	SLOG(LOG_DEBUG, TAG_TTSC, "Current state is 'Playing'");	break;
	case TTS_STATE_PAUSED:	SLOG(LOG_DEBUG, TAG_TTSC, "Current state is 'Paused'");		break;
	default:		SLOG(LOG_ERROR, TAG_TTSC, "[ERROR] Invalid value");		break;
	}

	return TTS_ERROR_NONE;
}

int tts_get_speed_range(tts_h tts, int* min, int* normal, int* max)
{
	if (0 != __tts_get_feature_enabled()) {
		return TTS_ERROR_NOT_SUPPORTED;
	}

	if (NULL == tts || NULL == min || NULL == normal || NULL == max) {
		SLOG(LOG_ERROR, TAG_TTSC, "[ERROR] Input parameter is null");
		return TTS_ERROR_INVALID_PARAMETER;
	}

	tts_client_s* client = tts_client_get(tts);

	if (NULL == client) {
		SLOG(LOG_ERROR, TAG_TTSC, "[ERROR] Get state : A handle is not valid");
		return TTS_ERROR_INVALID_PARAMETER;
	}

	*min = TTS_SPEED_MIN;
	*normal = TTS_SPEED_NORMAL;
	*max = TTS_SPEED_MAX;

	return TTS_ERROR_NONE;
}

int tts_get_error_message(tts_h tts, char** err_msg)
{
	if (0 != __tts_get_feature_enabled()) {
		return TTS_ERROR_NOT_SUPPORTED;
	}

	if (NULL == tts || NULL == err_msg) {
		SLOG(LOG_ERROR, TAG_TTSC, "[ERROR] Input parameter is null");
		return TTS_ERROR_INVALID_PARAMETER;
	}

	tts_client_s* client = tts_client_get(tts);

	if (NULL == client) {
		SLOG(LOG_ERROR, TAG_TTSC, "[ERROR] Get state : A handle is not valid");
		return TTS_ERROR_INVALID_PARAMETER;
	}

	if (NULL != client->err_msg) {
		*err_msg = strdup(client->err_msg);
		SLOG(LOG_DEBUG, TAG_TTSC, "[SUCCESS] Error msg (%s)", *err_msg);
	} else {
		*err_msg = NULL;
		SLOG(LOG_DEBUG, TAG_TTSC, "[SUCCESS] Error msg (NULL)");
	}

	SLOG(LOG_DEBUG, TAG_TTSC, "@@@");

	return TTS_ERROR_NONE;
}

int tts_add_text(tts_h tts, const char* text, const char* language, int voice_type, int speed, int* utt_id)
{
	SLOG(LOG_INFO, TAG_TTSC, "[DEBUG] Add text: text(%s), language(%s), type(%d)", (NULL == text) ? "NULL" : text, (NULL == language) ? "NULL" : language, voice_type);

	if (0 != __tts_get_feature_enabled()) {
		return TTS_ERROR_NOT_SUPPORTED;
	}

	if (speed < 0) {
		SLOG(LOG_ERROR, TAG_TTSC, "[ERROR] Speed should not be negative(%d)", speed);
		return TTS_ERROR_INVALID_PARAMETER;
	}

	if (voice_type < 0) {
		SLOG(LOG_ERROR, TAG_TTSC, "[ERROR] Voice type should not be negative(%d)", voice_type);
		return TTS_ERROR_INVALID_PARAMETER;
	}

	SLOG(LOG_DEBUG, TAG_TTSC, "@@@ Add text");

	if (NULL == tts || NULL == utt_id) {
		SLOG(LOG_ERROR, TAG_TTSC, "[ERROR] Input parameter is null");
		SLOG(LOG_DEBUG, TAG_TTSC, "@@@");
		return TTS_ERROR_INVALID_PARAMETER;
	}

	tts_client_s* client = tts_client_get(tts);

	if (NULL == client) {
		SLOG(LOG_ERROR, TAG_TTSC, "[ERROR] A handle is not valid");
		SLOG(LOG_DEBUG, TAG_TTSC, "@@@");
		return TTS_ERROR_INVALID_PARAMETER;
	}

	if (TTS_STATE_CREATED == client->current_state) {
		SLOG(LOG_ERROR, TAG_TTSC, "[ERROR] Current state is 'CREATED'.");
		return TTS_ERROR_INVALID_STATE;
	}

	if (-1 == g_max_text_size) {
		SLOG(LOG_DEBUG, TAG_TTSC, "[DEBUG] g_max_text_size is %d", g_max_text_size);
		if (0 != tts_config_mgr_get_max_text_size((unsigned int*)&g_max_text_size)) {
			SLOG(LOG_ERROR, TAG_TTSC, "[ERROR] Fail to get max text size");
			return TTS_ERROR_INVALID_PARAMETER;
		}
	}

	if (0 == g_max_text_size) {
		if (strlen(text) <= 0) {
			SLOG(LOG_DEBUG, TAG_TTSC, "[DEBUG] Max Text Size is %d", g_max_text_size);
			SLOG(LOG_ERROR, TAG_TTSC, "[ERROR] Input text size is invalid. (max text size is unlimited.)");
			SLOG(LOG_DEBUG, TAG_TTSC, "@@@");
			return TTS_ERROR_INVALID_PARAMETER;
		}
	} else {
		SLOG(LOG_DEBUG, TAG_TTSC, "[DEBUG] g_max_text_size is %d", g_max_text_size);
		if (g_max_text_size < strlen(text) || strlen(text) <= 0) {
			SLOG(LOG_ERROR, TAG_TTSC, "[ERROR] Input text size is invalid.");
			SLOG(LOG_DEBUG, TAG_TTSC, "@@@");
			return TTS_ERROR_INVALID_PARAMETER;
		}
	}

	if (TTS_SPEED_AUTO > speed || TTS_SPEED_MAX < speed) {
		SLOG(LOG_ERROR, TAG_TTSC, "[ERROR] speed value(%d) is invalid.", speed);
		SLOG(LOG_DEBUG, TAG_TTSC, "@@@");
		return TTS_ERROR_INVALID_PARAMETER;
	}

	if (false == g_screen_reader && TTS_MODE_SCREEN_READER == client->mode) {
		SLOG(LOG_WARN, TAG_TTSC, "[WARNING] Screen reader option is NOT available. Ignore this request");
		return TTS_ERROR_INVALID_STATE;
	}

	if (true == client->credential_needed && NULL == client->credential) {
		SLOG(LOG_ERROR, TAG_TTSC, "[ERROR] Do not have app credential for this engine");
		return TTS_ERROR_PERMISSION_DENIED;
	}

	/* check valid utf8 */
	bool valid = false;

	DBusError err;
	dbus_error_init(&err);

	valid = dbus_validate_utf8(text, &err);
	if (dbus_error_is_set(&err)) {
		SLOG(LOG_ERROR, TAG_TTSC, "[ERROR] Dbus Error(%s), text(%s)", err.message, text);
		dbus_error_free(&err);
		return TTS_ERROR_INVALID_PARAMETER;
	}

	if (valid != true) {
		SLOG(LOG_ERROR, TAG_TTSC, "[ERROR] Text is invalid - '%s'", text);
		return TTS_ERROR_INVALID_PARAMETER;
	}
	SLOG(LOG_DEBUG, TAG_TTSC, "Text is valid - text is '%s'", text);

	/* save texts for repetition */
	if (NULL != client->text_repeat) {
		free(client->text_repeat);
		client->text_repeat = NULL;
	}

	client->text_repeat = strdup(text);

	if (NULL != g_language) {
		free(g_language);
		g_language = NULL;
	}
	if (NULL == language)
		g_language = NULL;
	else
		g_language = strdup(language);

	g_voice_type = voice_type;
	g_speed = speed;

	SLOG(LOG_DEBUG, TAG_TTSC, "[DEBUG] repeat: text(%s), language(%s), voice type(%d), speed(%d)", client->text_repeat, g_language, g_voice_type, g_speed);

	/* change default language value */
	char* temp = NULL;

	if (NULL == language)
		temp = strdup("default");
	else
		temp = strdup(language);

	client->current_utt_id++;
	if (client->current_utt_id == 10000) {
		client->current_utt_id = 1;
	}

	/* do request */
	int ret = -1;
	int count = 0;
	bool is_prepared = false;
	while (0 != ret) {
		ret = tts_dbus_request_add_text(client->uid, text, temp, voice_type, speed, client->current_utt_id, client->credential);
		if (0 != ret) {
			if (TTS_ERROR_INVALID_PARAMETER == ret && false == is_prepared) {
				client->current_state = TTS_STATE_CREATED;
				if (0 == tts_prepare_sync(tts)) {
					is_prepared = true;
					SLOG(LOG_INFO, TAG_TTSC, "[INFO] Success tts_prepare_sync");
				}
			} else if (TTS_ERROR_TIMED_OUT != ret) {
				SLOG(LOG_ERROR, TAG_TTSC, "[ERROR] result : %s", __tts_get_error_code(ret));
				break;
			} else {
				SLOG(LOG_WARN, TAG_TTSC, "[WARNING] retry add text : %s", __tts_get_error_code(ret));
				usleep(10000);
				count++;
				if (TTS_RETRY_MIN_COUNT == count) {
					SLOG(LOG_ERROR, TAG_TTSC, "[ERROR] Fail to request");
					break;
				}
			}
		} else {
			*utt_id = client->current_utt_id;
		}
	}

	if (NULL != temp) {
		free(temp);
		temp = NULL;
	}

	SLOG(LOG_DEBUG, TAG_TTSC, "@@@");

	return ret;
}

static void __tts_play_async(void *data)
{
	tts_h tts = (tts_h)data;
	tts_client_s* client = tts_client_get(tts);

	/* check handle */
	if (NULL == client) {
		SLOG(LOG_ERROR, TAG_TTSC, "[ERROR] A handle is not valid");
		return;
	}

	int ret = -1;
	int count = 0;
	bool is_prepared = false;
	while (0 != ret) {
		ret = tts_dbus_request_play(client->uid, client->credential);
		if (0 != ret) {
			if (TTS_ERROR_INVALID_PARAMETER == ret && false == is_prepared) {
				client->current_state = TTS_STATE_CREATED;
				if (0 == tts_prepare_sync(tts)) {
					is_prepared = true;
					SLOG(LOG_INFO, TAG_TTSC, "[INFO] Success tts_prepare_sync");
				}
			} else if (TTS_ERROR_TIMED_OUT != ret) {
				SLOG(LOG_ERROR, TAG_TTSC, "[ERROR] result : %s", __tts_get_error_code(ret));
				break;
			} else {
				SLOG(LOG_WARN, TAG_TTSC, "[WARNING] retry play : %s", __tts_get_error_code(ret));
				usleep(10000);
				count++;
				if (TTS_RETRY_COUNT == count) {
					SLOG(LOG_ERROR, TAG_TTSC, "[ERROR] Fail to request");
					break;
				}
			}
		}
	}

	if (0 != ret) {
		SLOG(LOG_ERROR, TAG_TTSC, "[ERROR] Fail to play tts : %s", __tts_get_error_code(ret));

		client->reason = ret;
		client->utt_id = -1;

		ecore_timer_add(0, __tts_notify_error, client->tts);
		return;
	}

	client->before_state = client->current_state;
	client->current_state = TTS_STATE_PLAYING;

	if (NULL != client->state_changed_cb) {
		tts_client_use_callback(client);
		client->state_changed_cb(client->tts, client->before_state, client->current_state, client->state_changed_user_data);
		tts_client_not_use_callback(client);
		SLOG(LOG_DEBUG, TAG_TTSC, "State changed callback is called");
	}

	SLOG(LOG_DEBUG, TAG_TTSC, "@@@");

	return;
}

int tts_play_async(tts_h tts)
{
	if (0 != __tts_get_feature_enabled()) {
		return TTS_ERROR_NOT_SUPPORTED;
	}

	SLOG(LOG_DEBUG, TAG_TTSC, "@@@ Play tts");

	if (NULL == tts) {
		SLOG(LOG_ERROR, TAG_TTSC, "[ERROR] Input handle is null.");
		SLOG(LOG_DEBUG, TAG_TTSC, "@@@");
		return TTS_ERROR_INVALID_PARAMETER;
	}

	tts_client_s* client = tts_client_get(tts);

	if (NULL == client) {
		SLOG(LOG_ERROR, TAG_TTSC, "[ERROR] A handle is not valid.");
		SLOG(LOG_DEBUG, TAG_TTSC, "@@@");
		return TTS_ERROR_INVALID_PARAMETER;
	}

	if (TTS_STATE_PLAYING == client->current_state || TTS_STATE_CREATED == client->current_state) {
		SLOG(LOG_ERROR, TAG_TTSC, "[ERROR] The current state is invalid.");
		return TTS_ERROR_INVALID_STATE;
	}

	if (false == g_screen_reader && TTS_MODE_SCREEN_READER == client->mode) {
		SLOG(LOG_WARN, TAG_TTSC, "[WARNING] Screen reader option is NOT available. Ignore this request");
		return TTS_ERROR_INVALID_STATE;
	}

	if (true == client->credential_needed && NULL == client->credential) {
		SLOG(LOG_ERROR, TAG_TTSC, "[ERROR] Do not have app credential for this engine");
		return TTS_ERROR_PERMISSION_DENIED;
	}

	ecore_main_loop_thread_safe_call_async(__tts_play_async, (void*)tts);

	SLOG(LOG_DEBUG, TAG_TTSC, "@@@");

	return TTS_ERROR_NONE;
}

int tts_play(tts_h tts)
{
	if (0 != __tts_get_feature_enabled()) {
		return TTS_ERROR_NOT_SUPPORTED;
	}

	SLOG(LOG_INFO, TAG_TTSC, "@@@ Play tts");

	if (NULL == tts) {
		SLOG(LOG_ERROR, TAG_TTSC, "[ERROR] Input handle is null.");
		SLOG(LOG_DEBUG, TAG_TTSC, "@@@");
		return TTS_ERROR_INVALID_PARAMETER;
	}

	tts_client_s* client = tts_client_get(tts);

	if (NULL == client) {
		SLOG(LOG_ERROR, TAG_TTSC, "[ERROR] A handle is not valid.");
		SLOG(LOG_DEBUG, TAG_TTSC, "@@@");
		SLOG(LOG_DEBUG, TAG_TTSC, " ");
		return TTS_ERROR_INVALID_PARAMETER;
	}

	if (TTS_STATE_PLAYING == client->current_state || TTS_STATE_CREATED == client->current_state) {
		SLOG(LOG_ERROR, TAG_TTSC, "[ERROR] The current state is invalid.");
		return TTS_ERROR_INVALID_STATE;
	}

	if (false == g_screen_reader && TTS_MODE_SCREEN_READER == client->mode) {
		SLOG(LOG_WARN, TAG_TTSC, "[WARNING] Screen reader option is NOT available. Ignore this request");
		return TTS_ERROR_INVALID_STATE;
	}

	if (true == client->credential_needed && NULL == client->credential) {
		SLOG(LOG_ERROR, TAG_TTSC, "[ERROR] Do not have app credential for this engine");
		return TTS_ERROR_PERMISSION_DENIED;
	}

	int ret = -1;
	int count = 0;
	bool is_prepared = false;
	while (0 != ret) {
		ret = tts_dbus_request_play(client->uid, client->credential);
		if (0 != ret) {
			if (TTS_ERROR_INVALID_PARAMETER == ret && false == is_prepared) {
				client->current_state = TTS_STATE_CREATED;
				if (0 == tts_prepare_sync(tts)) {
					is_prepared = true;
					SLOG(LOG_INFO, TAG_TTSC, "[INFO] Success tts_prepare_sync");
				}
			} else if (TTS_ERROR_TIMED_OUT != ret) {
				SLOG(LOG_ERROR, TAG_TTSC, "[ERROR] result : %s", __tts_get_error_code(ret));
				return ret;
			} else {
				SLOG(LOG_WARN, TAG_TTSC, "[WARNING] retry play : %s", __tts_get_error_code(ret));
				usleep(10000);
				count++;
				if (TTS_RETRY_COUNT == count) {
					SLOG(LOG_ERROR, TAG_TTSC, "[ERROR] Fail to request");
					return ret;
				}
			}
		}
	}

	client->before_state = client->current_state;
	client->current_state = TTS_STATE_PLAYING;

	if (NULL != client->state_changed_cb) {
		tts_client_use_callback(client);
		client->state_changed_cb(client->tts, client->before_state, client->current_state, client->state_changed_user_data);
		tts_client_not_use_callback(client);
		SLOG(LOG_DEBUG, TAG_TTSC, "State changed callback is called");
	}

	SLOG(LOG_DEBUG, TAG_TTSC, "@@@");

	return TTS_ERROR_NONE;
}

static void __tts_stop_async(void *data)
{
	tts_h tts = (tts_h)data;
	tts_client_s* client = tts_client_get(tts);

	/* check handle */
	if (NULL == client) {
		SLOG(LOG_ERROR, TAG_TTSC, "[ERROR] A handle is not valid");
		return;
	}

	int ret = -1;
	int count = 0;
	bool is_prepared = false;
	while (0 != ret) {
		ret = tts_dbus_request_stop(client->uid);
		if (0 != ret) {
			if (TTS_ERROR_INVALID_PARAMETER == ret && false == is_prepared) {
				client->current_state = TTS_STATE_CREATED;
				if (0 == tts_prepare_sync(tts)) {
					is_prepared = true;
					SLOG(LOG_INFO, TAG_TTSC, "[INFO] Success tts_prepare_sync");
				}
			} else if (TTS_ERROR_TIMED_OUT != ret) {
				SLOG(LOG_ERROR, TAG_TTSC, "[ERROR] result : %s", __tts_get_error_code(ret));
				break;
			} else {
				SLOG(LOG_WARN, TAG_TTSC, "[WARNING] retry stop : %s", __tts_get_error_code(ret));
				usleep(10000);
				count++;
				if (TTS_RETRY_COUNT == count) {
					SLOG(LOG_ERROR, TAG_TTSC, "[ERROR] Fail to request");
					break;
				}
			}
		}
	}

	if (0 != ret) {
		SLOG(LOG_ERROR, TAG_TTSC, "[ERROR] Fail to stop tts : %s", __tts_get_error_code(ret));

		client->reason = ret;
		client->utt_id = -1;

		ecore_timer_add(0, __tts_notify_error, client->tts);
		return;
	}

	client->before_state = client->current_state;
	client->current_state = TTS_STATE_READY;

	if (NULL != client->state_changed_cb) {
		tts_client_use_callback(client);
		client->state_changed_cb(client->tts, client->before_state, client->current_state, client->state_changed_user_data);
		tts_client_not_use_callback(client);
		SLOG(LOG_DEBUG, TAG_TTSC, "State changed callback is called");
	}

	SLOG(LOG_DEBUG, TAG_TTSC, "@@@");

	return;
}

int tts_stop_aync(tts_h tts)
{
	if (0 != __tts_get_feature_enabled()) {
		return TTS_ERROR_NOT_SUPPORTED;
	}

	SLOG(LOG_DEBUG, TAG_TTSC, "@@@ Stop tts");

	if (NULL == tts) {
		SLOG(LOG_ERROR, TAG_TTSC, "[ERROR] Input handle is null");
		SLOG(LOG_DEBUG, TAG_TTSC, "@@@");
		return TTS_ERROR_INVALID_PARAMETER;
	}

	tts_client_s* client = tts_client_get(tts);

	if (NULL == client) {
		SLOG(LOG_ERROR, TAG_TTSC, "[ERROR] A handle is not valid");
		SLOG(LOG_DEBUG, TAG_TTSC, "@@@");
		return TTS_ERROR_INVALID_PARAMETER;
	}

	if (TTS_STATE_CREATED == client->current_state) {
		SLOG(LOG_ERROR, TAG_TTSC, "[ERROR] Current state is 'CREATED'.");
		return TTS_ERROR_INVALID_STATE;
	}

	if (false == g_screen_reader && TTS_MODE_SCREEN_READER == client->mode) {
		SLOG(LOG_WARN, TAG_TTSC, "[WARNING] Screen reader option is NOT available. Ignore this request");
		return TTS_ERROR_INVALID_STATE;
	}

	ecore_main_loop_thread_safe_call_async(__tts_stop_async, (void*)tts);

	SLOG(LOG_DEBUG, TAG_TTSC, "@@@");

	return TTS_ERROR_NONE;
}

int tts_stop(tts_h tts)
{
	if (0 != __tts_get_feature_enabled()) {
		return TTS_ERROR_NOT_SUPPORTED;
	}

	SLOG(LOG_INFO, TAG_TTSC, "@@@ Stop tts");

	if (NULL == tts) {
		SLOG(LOG_ERROR, TAG_TTSC, "[ERROR] Input handle is null");
		SLOG(LOG_DEBUG, TAG_TTSC, "@@@");
		return TTS_ERROR_INVALID_PARAMETER;
	}

	tts_client_s* client = tts_client_get(tts);

	if (NULL == client) {
		SLOG(LOG_ERROR, TAG_TTSC, "[ERROR] A handle is not valid");
		SLOG(LOG_DEBUG, TAG_TTSC, "@@@");
		return TTS_ERROR_INVALID_PARAMETER;
	}

	if (TTS_STATE_CREATED == client->current_state) {
		SLOG(LOG_ERROR, TAG_TTSC, "[ERROR] Current state is 'CREATED'.");
		return TTS_ERROR_INVALID_STATE;
	}

	if (false == g_screen_reader && TTS_MODE_SCREEN_READER == client->mode) {
		SLOG(LOG_WARN, TAG_TTSC, "[WARNING] Screen reader option is NOT available. Ignore this request");
		return TTS_ERROR_INVALID_STATE;
	}

	int ret = -1;
	int count = 0;
	bool is_prepared = false;
	while (0 != ret) {
		ret = tts_dbus_request_stop(client->uid);
		if (0 != ret) {
			if (TTS_ERROR_INVALID_PARAMETER == ret && false == is_prepared) {
				client->current_state = TTS_STATE_CREATED;
				if (0 == tts_prepare_sync(tts)) {
					is_prepared = true;
					SLOG(LOG_INFO, TAG_TTSC, "[INFO] Success tts_prepare_sync");
				}
			} else if (TTS_ERROR_TIMED_OUT != ret) {
				SLOG(LOG_ERROR, TAG_TTSC, "[ERROR] result : %s", __tts_get_error_code(ret));
				return ret;
			} else {
				SLOG(LOG_WARN, TAG_TTSC, "[WARNING] retry stop : %s", __tts_get_error_code(ret));
				usleep(10000);
				count++;
				if (TTS_RETRY_COUNT == count) {
					SLOG(LOG_ERROR, TAG_TTSC, "[ERROR] Fail to request");
					return ret;
				}
			}
		}
	}

	client->before_state = client->current_state;
	client->current_state = TTS_STATE_READY;

	if (NULL != client->state_changed_cb) {
		tts_client_use_callback(client);
		client->state_changed_cb(client->tts, client->before_state, client->current_state, client->state_changed_user_data);
		tts_client_not_use_callback(client);
		SLOG(LOG_DEBUG, TAG_TTSC, "State changed callback is called");
	}

	SLOG(LOG_DEBUG, TAG_TTSC, "@@@");

	return TTS_ERROR_NONE;
}

static void __tts_pause_async(void *data)
{
	tts_h tts = (tts_h)data;
	tts_client_s* client = tts_client_get(tts);

	/* check handle */
	if (NULL == client) {
		SLOG(LOG_ERROR, TAG_TTSC, "[ERROR] A handle is not valid");
		return;
	}

	int ret = -1;
	int count = 0;
	bool is_prepared = false;
	while (0 != ret) {
		ret = tts_dbus_request_pause(client->uid);
		if (0 != ret) {
			if (TTS_ERROR_INVALID_PARAMETER == ret && false == is_prepared) {
				client->current_state = TTS_STATE_CREATED;
				if (0 == tts_prepare_sync(tts)) {
					is_prepared = true;
					SLOG(LOG_INFO, TAG_TTSC, "[INFO] Success tts_prepare_sync");
				}
			} else if (TTS_ERROR_TIMED_OUT != ret) {
				SLOG(LOG_ERROR, TAG_TTSC, "[ERROR] result : %s", __tts_get_error_code(ret));
				break;
			} else {
				SLOG(LOG_WARN, TAG_TTSC, "[WARNING] retry pause : %s", __tts_get_error_code(ret));
				usleep(10000);
				count++;
				if (TTS_RETRY_COUNT == count) {
					SLOG(LOG_ERROR, TAG_TTSC, "[ERROR] Fail to request");
					break;
				}
			}
		}
	}

	if (0 != ret) {
		SLOG(LOG_ERROR, TAG_TTSC, "[ERROR] Fail to pause tts : %s", __tts_get_error_code(ret));

		client->reason = ret;
		client->utt_id = -1;

		ecore_timer_add(0, __tts_notify_error, client->tts);
		return;
	}

	client->before_state = client->current_state;
	client->current_state = TTS_STATE_PAUSED;

	if (NULL != client->state_changed_cb) {
		tts_client_use_callback(client);
		client->state_changed_cb(client->tts, client->before_state, client->current_state, client->state_changed_user_data);
		tts_client_not_use_callback(client);
		SLOG(LOG_DEBUG, TAG_TTSC, "State changed callback is called");
	}

	SLOG(LOG_DEBUG, TAG_TTSC, "@@@");

	return;
}

int tts_pause_async(tts_h tts)
{
	if (0 != __tts_get_feature_enabled()) {
		return TTS_ERROR_NOT_SUPPORTED;
	}

	SLOG(LOG_DEBUG, TAG_TTSC, "@@@ Pause tts");

	if (NULL == tts) {
		SLOG(LOG_ERROR, TAG_TTSC, "[ERROR] Input handle is null");
		SLOG(LOG_DEBUG, TAG_TTSC, "@@@");
		return TTS_ERROR_INVALID_PARAMETER;
	}

	tts_client_s* client = tts_client_get(tts);

	if (NULL == client) {
		SLOG(LOG_ERROR, TAG_TTSC, "[ERROR] A handle is not valid");
		SLOG(LOG_DEBUG, TAG_TTSC, "@@@");
		return TTS_ERROR_INVALID_PARAMETER;
	}

	if (TTS_STATE_PLAYING != client->current_state) {
		SLOG(LOG_ERROR, TAG_TTSC, "[Error] The Current state is NOT 'playing'. So this request should be not running.");
		SLOG(LOG_DEBUG, TAG_TTSC, "@@@");
		return TTS_ERROR_INVALID_STATE;
	}

	if (false == g_screen_reader && TTS_MODE_SCREEN_READER == client->mode) {
		SLOG(LOG_WARN, TAG_TTSC, "[WARNING] Screen reader option is NOT available. Ignore this request");
		return TTS_ERROR_INVALID_STATE;
	}

	ecore_main_loop_thread_safe_call_async(__tts_pause_async, (void*)tts);

	SLOG(LOG_DEBUG, TAG_TTSC, "@@@");

	return TTS_ERROR_NONE;
}

int tts_pause(tts_h tts)
{
	if (0 != __tts_get_feature_enabled()) {
		return TTS_ERROR_NOT_SUPPORTED;
	}

	SLOG(LOG_INFO, TAG_TTSC, "@@@ Pause tts");

	if (NULL == tts) {
		SLOG(LOG_ERROR, TAG_TTSC, "[ERROR] Input handle is null");
		SLOG(LOG_DEBUG, TAG_TTSC, "@@@");
		return TTS_ERROR_INVALID_PARAMETER;
	}

	tts_client_s* client = tts_client_get(tts);

	if (NULL == client) {
		SLOG(LOG_ERROR, TAG_TTSC, "[ERROR] A handle is not valid");
		SLOG(LOG_DEBUG, TAG_TTSC, "@@@");
		return TTS_ERROR_INVALID_PARAMETER;
	}

	if (TTS_STATE_PLAYING != client->current_state) {
		SLOG(LOG_ERROR, TAG_TTSC, "[Error] The Current state is NOT 'playing'. So this request should be not running.");
		SLOG(LOG_DEBUG, TAG_TTSC, "@@@");
		return TTS_ERROR_INVALID_STATE;
	}

	if (false == g_screen_reader && TTS_MODE_SCREEN_READER == client->mode) {
		SLOG(LOG_WARN, TAG_TTSC, "[WARNING] Screen reader option is NOT available. Ignore this request");
		return TTS_ERROR_INVALID_STATE;
	}

	int ret = -1;
	int count = 0;
	bool is_prepared = false;
	while (0 != ret) {
		ret = tts_dbus_request_pause(client->uid);
		if (0 != ret) {
			if (TTS_ERROR_INVALID_PARAMETER == ret && false == is_prepared) {
				client->current_state = TTS_STATE_CREATED;
				if (0 == tts_prepare_sync(tts)) {
					is_prepared = true;
					SLOG(LOG_INFO, TAG_TTSC, "[INFO] Success tts_prepare_sync");
				}
			} else if (TTS_ERROR_TIMED_OUT != ret) {
				SLOG(LOG_ERROR, TAG_TTSC, "[ERROR] result : %s", __tts_get_error_code(ret));
				return ret;
			} else {
				SLOG(LOG_WARN, TAG_TTSC, "[WARNING] retry pause : %s", __tts_get_error_code(ret));
				usleep(10000);
				count++;
				if (TTS_RETRY_COUNT == count) {
					SLOG(LOG_ERROR, TAG_TTSC, "[ERROR] Fail to request");
					return ret;
				}
			}
		}
	}

	client->before_state = client->current_state;
	client->current_state = TTS_STATE_PAUSED;

	if (NULL != client->state_changed_cb) {
		tts_client_use_callback(client);
		client->state_changed_cb(client->tts, client->before_state, client->current_state, client->state_changed_user_data);
		tts_client_not_use_callback(client);
		SLOG(LOG_DEBUG, TAG_TTSC, "State changed callback is called");
	}

	SLOG(LOG_DEBUG, TAG_TTSC, "@@@");

	return TTS_ERROR_NONE;
}

int tts_set_private_data(tts_h tts, const char* key, const char* data)
{
	if (0 != __tts_get_feature_enabled()) {
		return TTS_ERROR_NOT_SUPPORTED;
	}

	SLOG(LOG_INFO, TAG_TTSC, "@@@ Set private data, key(%s), data(%s)", key, data);

	if (NULL == tts) {
		SLOG(LOG_ERROR, TAG_TTSC, "[ERROR] Input handle isnull");
		return TTS_ERROR_INVALID_PARAMETER;
	}

	if (NULL == key || NULL == data) {
		SLOG(LOG_ERROR, TAG_TTSC, "[ERROR] Invalid parameter");
		return TTS_ERROR_INVALID_PARAMETER;
	}

	tts_client_s* client = tts_client_get(tts);

	if (NULL == client) {
		SLOG(LOG_ERROR, TAG_TTSC, "[ERROR] A handle is not valid");
		return TTS_ERROR_INVALID_PARAMETER;
	}

	if (TTS_STATE_READY != client->current_state) {
		SLOG(LOG_ERROR, TAG_TTSC, "[ERROR] Invalid state : Current state is NOT 'READY'");
		return TTS_ERROR_INVALID_STATE;
	}

	if (true != client->internal && (0 == strcmp(key, "EnableServerTTS") || 0 == strcmp(key, "DisableServerTTS"))) {
		SLOG(LOG_ERROR, TAG_TTSC, "[ERROR] This is not an internal app");
		return TTS_ERROR_INVALID_PARAMETER;
	}

	int ret = -1;
	int count = 0;
	bool is_prepared = false;
	while (0 != ret) {
		ret = tts_dbus_request_set_private_data(client->uid, key, data);
		if (0 != ret) {
			if (TTS_ERROR_INVALID_PARAMETER == ret && false == is_prepared) {
				client->current_state = TTS_STATE_CREATED;
				if (0 == tts_prepare_sync(tts)) {
					is_prepared = true;
					SLOG(LOG_INFO, TAG_TTSC, "[INFO] Success tts_prepare_sync");
				}
			} else if (TTS_ERROR_TIMED_OUT != ret) {
				SLOG(LOG_ERROR, TAG_TTSC, "[ERROR] result : %s", __tts_get_error_code(ret));
				return ret;
			} else {
				SLOG(LOG_WARN, TAG_TTSC, "[WARNING] retry : %s", __tts_get_error_code(ret));
				usleep(10000);
				count++;
				if (TTS_RETRY_COUNT == count) {
					SLOG(LOG_ERROR, TAG_TTSC, "[ERROR] Fail to request");
					return ret;
				}
			}
		}
	}

	SLOG(LOG_DEBUG, TAG_TTSC, "@@@");

	return 0;
}

int tts_get_private_data(tts_h tts, const char* key, char** data)
{
	if (0 != __tts_get_feature_enabled()) {
		return TTS_ERROR_NOT_SUPPORTED;
	}

	SLOG(LOG_INFO, TAG_TTSC, "@@@ Get private data, key(%s)", key);

	if (NULL == tts) {
		SLOG(LOG_ERROR, TAG_TTSC, "[ERROR] Input handle is null");
		return TTS_ERROR_INVALID_PARAMETER;
	}

	if (NULL == key || NULL == data) {
		SLOG(LOG_ERROR, TAG_TTSC, "[ERROR] Invalid parameter");
		return TTS_ERROR_INVALID_PARAMETER;
	}

	tts_client_s* client = tts_client_get(tts);

	if (NULL == client) {
		SLOG(LOG_ERROR, TAG_TTSC, "[ERROR] A handle is not valid");
		return TTS_ERROR_INVALID_PARAMETER;
	}

	if (TTS_STATE_READY != client->current_state) {
		SLOG(LOG_ERROR, TAG_TTSC, "[ERROR] Invalid state : Current state is NOT 'READY'");
		return TTS_ERROR_INVALID_STATE;
	}

	int ret = -1;
	int count = 0;
	bool is_prepared = false;
	while (0 != ret) {
		ret = tts_dbus_request_get_private_data(client->uid, key, data);
		if (0 != ret) {
			if (TTS_ERROR_INVALID_PARAMETER == ret && false == is_prepared) {
				client->current_state = TTS_STATE_CREATED;
				if (0 == tts_prepare_sync(tts)) {
					is_prepared = true;
					SLOG(LOG_INFO, TAG_TTSC, "[INFO] Success tts_prepare_sync");
				}
			} else if (TTS_ERROR_TIMED_OUT != ret) {
				SLOG(LOG_ERROR, TAG_TTSC, "[ERROR] result : %s", __tts_get_error_code(ret));
				return ret;
			} else {
				SLOG(LOG_WARN, TAG_TTSC, "[WARNING] retry : %s", __tts_get_error_code(ret));
				usleep(10000);
				count++;
				if (TTS_RETRY_COUNT == count) {
					SLOG(LOG_ERROR, TAG_TTSC, "[ERROR] Fail to request");
					return ret;
				}
			}
		}
	}

	if (0 == strncmp(*data, "NULL", strlen(*data))) {
		free(*data);
		*data = NULL;
	}

	SLOG(LOG_DEBUG, TAG_TTSC, "@@@");

	return 0;
}

static Eina_Bool __tts_notify_error(void *data)
{
	tts_h tts = (tts_h)data;

	tts_client_s* client = tts_client_get(tts);

	/* check handle */
	if (NULL == client) {
		SLOG(LOG_WARN, TAG_TTSC, "Fail to notify error msg : A handle is not valid");
		return EINA_FALSE;
	}

	SLOG(LOG_DEBUG, TAG_TTSC, "Error data : uttid(%d) reason(%s)", client->utt_id, __tts_get_error_code(client->reason));

	if (NULL != client->error_cb) {
		SLOG(LOG_DEBUG, TAG_TTSC, "Call callback function of error");
		tts_client_use_callback(client);
		g_err_callback_status = true;
		client->error_cb(client->tts, client->utt_id, client->reason, client->error_user_data);
		g_err_callback_status = false;
		tts_client_not_use_callback(client);
	} else {
		SLOG(LOG_WARN, TAG_TTSC, "No registered callback function of error ");
	}

	return EINA_FALSE;
}

int __tts_cb_error(int uid, tts_error_e reason, int utt_id, char* err_msg)
{
	if (-1 == uid) {
		GList* client_list = NULL;
		client_list = tts_client_get_client_list();

		GList *iter = NULL;
		tts_client_s *data = NULL;

		if (g_list_length(client_list) > 0) {
			/* Get a first item */
			iter = g_list_first(client_list);

			while (NULL != iter) {
				data = iter->data;

				data->utt_id = utt_id;
				data->reason = reason;
				if (NULL != data->err_msg) {
					free(data->err_msg);
					data->err_msg = NULL;
				}
				if (NULL != err_msg)
					data->err_msg = strdup(err_msg);

				/* call callback function */
				if (NULL != data->error_cb) {
					ecore_timer_add(0, __tts_notify_error, data->tts);
				} else {
					SLOG(LOG_WARN, TAG_TTSC, "No registered callback function of error ");
				}

				if (TTS_ERROR_SERVICE_RESET == reason) {
					SLOG(LOG_WARN, TAG_TTSC, "[WARNING] Service Reset");

					data->current_state = TTS_STATE_CREATED;
					if (0 != tts_prepare(data->tts)) {
						SLOG(LOG_ERROR, TAG_TTSC, "[ERROR] Fail to prepare");
					}
				}

				/* Next item */
				iter = g_list_next(iter);
			}
		}
	} else {
		tts_client_s* client = tts_client_get_by_uid(uid);

		if (NULL == client) {
			SLOG(LOG_WARN, TAG_TTSC, "[WARNING] A handle is not valid");
			return TTS_ERROR_INVALID_PARAMETER;
		}

		client->utt_id = utt_id;
		client->reason = reason;
		if (NULL != client->err_msg) {
			free(client->err_msg);
			client->err_msg = NULL;
		}
		if (NULL != err_msg)
			client->err_msg = strdup(err_msg);

		/* call callback function */
		if (NULL != client->error_cb) {
			ecore_timer_add(0, __tts_notify_error, client->tts);
		} else {
			SLOG(LOG_WARN, TAG_TTSC, "No registered callback function of error ");
		}

		if (TTS_ERROR_SERVICE_RESET == reason) {
			SLOG(LOG_WARN, TAG_TTSC, "[WARNING] Service Reset");

			client->current_state = TTS_STATE_CREATED;
			if (0 != tts_prepare(client->tts)) {
				SLOG(LOG_ERROR, TAG_TTSC, "[ERROR] Fail to prepare");
			}
		}
	}

	return 0;
}

static Eina_Bool __tts_notify_state_changed(void *data)
{
	tts_h tts = (tts_h)data;

	tts_client_s* client = tts_client_get(tts);

	/* check handle */
	if (NULL == client) {
		SLOG(LOG_WARN, TAG_TTSC, "Fail to notify state changed : A handle is not valid");
		return EINA_FALSE;
	}

	if (NULL != client->state_changed_cb) {
		tts_client_use_callback(client);
		client->state_changed_cb(client->tts, client->before_state, client->current_state, client->state_changed_user_data);
		tts_client_not_use_callback(client);
		SLOG(LOG_DEBUG, TAG_TTSC, "State changed callback is called : pre(%d) cur(%d)", client->before_state, client->current_state);
	} else {
		SLOG(LOG_WARN, TAG_TTSC, "[WARNING] State changed callback is null");
	}

	return EINA_FALSE;
}

int __tts_cb_set_state(int uid, int state)
{
	tts_client_s* client = tts_client_get_by_uid(uid);
	if (NULL == client) {
		SLOG(LOG_WARN, TAG_TTSC, "[WARNING] The handle is not valid");
		return -1;
	}

	tts_state_e state_from_daemon = (tts_state_e)state;

	if (client->current_state == state_from_daemon) {
		SLOG(LOG_DEBUG, TAG_TTSC, "Current state has already been %d", client->current_state);
		return 0;
	}

	if (NULL != client->state_changed_cb) {
		ecore_timer_add(0, __tts_notify_state_changed, client->tts);
	} else {
		SLOG(LOG_WARN, TAG_TTSC, "[WARNING] State changed callback is null");
	}

	client->before_state = client->current_state;
	client->current_state = state_from_daemon;

	return 0;
}

int __tts_cb_utt_started(int uid, int utt_id)
{
	tts_client_s* client = tts_client_get_by_uid(uid);

	if (NULL == client) {
		SLOG(LOG_WARN, TAG_TTSC, "[WARNING] A handle is not valid");
		return TTS_ERROR_INVALID_PARAMETER;
	}

	SLOG(LOG_INFO, TAG_TTSC, "utterance started : utt id(%d) ", utt_id);

	client->utt_id = utt_id;

	/* call callback function */
	if (NULL != client->utt_started_cb) {
		SLOG(LOG_DEBUG, TAG_TTSC, "Call callback function of utterance started ");
		tts_client_use_callback(client);
		client->utt_started_cb(client->tts, client->utt_id, client->utt_started_user_data);
		tts_client_not_use_callback(client);
	} else {
		SLOG(LOG_WARN, TAG_TTSC, "No registered callback function of utterance started ");
	}

	return 0;
}

int __tts_cb_utt_completed(int uid, int utt_id)
{
	tts_client_s* client = tts_client_get_by_uid(uid);

	if (NULL == client) {
		SLOG(LOG_WARN, TAG_TTSC, "[WARNING] A handle is not valid");
		return TTS_ERROR_INVALID_PARAMETER;
	}

	SLOG(LOG_INFO, TAG_TTSC, "utterance completed : uttid(%d) ", utt_id);

	client->utt_id = utt_id;

	/* call callback function */
	if (NULL != client->utt_completeted_cb) {
		SLOG(LOG_DEBUG, TAG_TTSC, "Call callback function of utterance completed ");
		tts_client_use_callback(client);
		client->utt_completeted_cb(client->tts, client->utt_id, client->utt_completed_user_data);
		tts_client_not_use_callback(client);
	} else {
		SLOG(LOG_WARN, TAG_TTSC, "No registered callback function of utterance completed ");
	}

	return 0;
}

int tts_set_state_changed_cb(tts_h tts, tts_state_changed_cb callback, void* user_data)
{
	if (0 != __tts_get_feature_enabled()) {
		return TTS_ERROR_NOT_SUPPORTED;
	}

	if (NULL == tts || NULL == callback) {
		SLOG(LOG_ERROR, TAG_TTSC, "[ERROR] Set state changed cb : Input parameter is null");
		return TTS_ERROR_INVALID_PARAMETER;
	}

	tts_client_s* client = tts_client_get(tts);

	if (NULL == client) {
		SLOG(LOG_ERROR, TAG_TTSC, "[ERROR] Set state changed cb : A handle is not valid");
		return TTS_ERROR_INVALID_PARAMETER;
	}

	if (TTS_STATE_CREATED != client->current_state) {
		SLOG(LOG_ERROR, TAG_TTSC, "[ERROR] Set state changed cb : Current state is not 'Created'.");
		return TTS_ERROR_INVALID_STATE;
	}

	client->state_changed_cb = callback;
	client->state_changed_user_data = user_data;

	SLOG(LOG_INFO, TAG_TTSC, "[SUCCESS] Set state changed cb");

	return 0;
}

int tts_unset_state_changed_cb(tts_h tts)
{
	if (0 != __tts_get_feature_enabled()) {
		return TTS_ERROR_NOT_SUPPORTED;
	}

	if (NULL == tts) {
		SLOG(LOG_ERROR, TAG_TTSC, "[ERROR] Unset state changed cb : Input parameter is null");
		return TTS_ERROR_INVALID_PARAMETER;
	}

	tts_client_s* client = tts_client_get(tts);

	if (NULL == client) {
		SLOG(LOG_ERROR, TAG_TTSC, "[ERROR] Unset state changed cb : A handle is not valid");
		return TTS_ERROR_INVALID_PARAMETER;
	}

	if (TTS_STATE_CREATED != client->current_state) {
		SLOG(LOG_ERROR, TAG_TTSC, "[ERROR] Unset state changed cb : Current state is not 'Created'.");
		return TTS_ERROR_INVALID_STATE;
	}

	client->state_changed_cb = NULL;
	client->state_changed_user_data = NULL;

	SLOG(LOG_DEBUG, TAG_TTSC, "[SUCCESS] Unset state changed cb");

	return 0;
}

int tts_set_utterance_started_cb(tts_h tts, tts_utterance_started_cb callback, void* user_data)
{
	if (0 != __tts_get_feature_enabled()) {
		return TTS_ERROR_NOT_SUPPORTED;
	}

	if (NULL == tts || NULL == callback) {
		SLOG(LOG_ERROR, TAG_TTSC, "[ERROR] Set utt started cb : Input parameter is null");
		return TTS_ERROR_INVALID_PARAMETER;
	}

	tts_client_s* client = tts_client_get(tts);

	if (NULL == client) {
		SLOG(LOG_ERROR, TAG_TTSC, "[ERROR] Set utt started cb : A handle is not valid");
		return TTS_ERROR_INVALID_PARAMETER;
	}

	if (TTS_STATE_CREATED != client->current_state) {
		SLOG(LOG_ERROR, TAG_TTSC, "[ERROR] Set utt started cb : Current state is not 'Created'.");
		return TTS_ERROR_INVALID_STATE;
	}

	client->utt_started_cb = callback;
	client->utt_started_user_data = user_data;

	SLOG(LOG_DEBUG, TAG_TTSC, "[SUCCESS] Set utt started cb");

	return 0;
}

int tts_unset_utterance_started_cb(tts_h tts)
{
	if (0 != __tts_get_feature_enabled()) {
		return TTS_ERROR_NOT_SUPPORTED;
	}

	if (NULL == tts) {
		SLOG(LOG_ERROR, TAG_TTSC, "[ERROR] Unset utt started cb : Input parameter is null");
		return TTS_ERROR_INVALID_PARAMETER;
	}

	tts_client_s* client = tts_client_get(tts);

	if (NULL == client) {
		SLOG(LOG_ERROR, TAG_TTSC, "[ERROR] Unset utt started cb : A handle is not valid");
		return TTS_ERROR_INVALID_PARAMETER;
	}

	if (TTS_STATE_CREATED != client->current_state) {
		SLOG(LOG_ERROR, TAG_TTSC, "[ERROR] Unset utt started cb : Current state is not 'Created'.");
		return TTS_ERROR_INVALID_STATE;
	}

	client->utt_started_cb = NULL;
	client->utt_started_user_data = NULL;

	SLOG(LOG_DEBUG, TAG_TTSC, "[SUCCESS] Unset utt started cb");

	return 0;
}

int tts_set_utterance_completed_cb(tts_h tts, tts_utterance_completed_cb callback, void* user_data)
{
	if (0 != __tts_get_feature_enabled()) {
		return TTS_ERROR_NOT_SUPPORTED;
	}

	if (NULL == tts || NULL == callback) {
		SLOG(LOG_ERROR, TAG_TTSC, "[ERROR] Set utt completed cb : Input parameter is null");
		return TTS_ERROR_INVALID_PARAMETER;
	}

	tts_client_s* client = tts_client_get(tts);

	if (NULL == client) {
		SLOG(LOG_ERROR, TAG_TTSC, "[ERROR] Set utt completed cb : A handle is not valid");
		return TTS_ERROR_INVALID_PARAMETER;
	}

	if (TTS_STATE_CREATED != client->current_state) {
		SLOG(LOG_ERROR, TAG_TTSC, "[ERROR] Set utt completed cb : Current state is not 'Created'.");
		return TTS_ERROR_INVALID_STATE;
	}

	client->utt_completeted_cb = callback;
	client->utt_completed_user_data = user_data;

	SLOG(LOG_DEBUG, TAG_TTSC, "[SUCCESS] Set utt completed cb");

	return 0;
}

int tts_unset_utterance_completed_cb(tts_h tts)
{
	if (0 != __tts_get_feature_enabled()) {
		return TTS_ERROR_NOT_SUPPORTED;
	}

	if (NULL == tts) {
		SLOG(LOG_ERROR, TAG_TTSC, "[ERROR] Unset utt completed cb : Input parameter is null");
		return TTS_ERROR_INVALID_PARAMETER;
	}

	tts_client_s* client = tts_client_get(tts);

	if (NULL == client) {
		SLOG(LOG_ERROR, TAG_TTSC, "[ERROR] Unset utt completed cb : A handle is not valid");
		return TTS_ERROR_INVALID_PARAMETER;
	}

	if (TTS_STATE_CREATED != client->current_state) {
		SLOG(LOG_ERROR, TAG_TTSC, "[ERROR] Unset utt completed cb : Current state is not 'Created'.");
		return TTS_ERROR_INVALID_STATE;
	}

	client->utt_completeted_cb = NULL;
	client->utt_completed_user_data = NULL;

	SLOG(LOG_DEBUG, TAG_TTSC, "[SUCCESS] Unset utt completed cb");
	return 0;
}

int tts_set_error_cb(tts_h tts, tts_error_cb callback, void* user_data)
{
	if (0 != __tts_get_feature_enabled()) {
		return TTS_ERROR_NOT_SUPPORTED;
	}

	if (NULL == tts || NULL == callback) {
		SLOG(LOG_ERROR, TAG_TTSC, "[ERROR] Set error cb : Input parameter is null");
		return TTS_ERROR_INVALID_PARAMETER;
	}

	tts_client_s* client = tts_client_get(tts);

	if (NULL == client) {
		SLOG(LOG_ERROR, TAG_TTSC, "[ERROR] Set error cb : A handle is not valid");
		return TTS_ERROR_INVALID_PARAMETER;
	}

	if (TTS_STATE_CREATED != client->current_state) {
		SLOG(LOG_ERROR, TAG_TTSC, "[ERROR] Set error cb : Current state is not 'Created'.");
		return TTS_ERROR_INVALID_STATE;
	}

	client->error_cb = callback;
	client->error_user_data = user_data;

	SLOG(LOG_DEBUG, TAG_TTSC, "[SUCCESS] Set error cb");

	return 0;
}

int tts_unset_error_cb(tts_h tts)
{
	if (0 != __tts_get_feature_enabled()) {
		return TTS_ERROR_NOT_SUPPORTED;
	}

	if (NULL == tts) {
		SLOG(LOG_ERROR, TAG_TTSC, "[ERROR] Unset error cb : Input parameter is null");
		return TTS_ERROR_INVALID_PARAMETER;
	}

	tts_client_s* client = tts_client_get(tts);

	if (NULL == client) {
		SLOG(LOG_ERROR, TAG_TTSC, "[ERROR] Unset error cb : A handle is not valid");
		return TTS_ERROR_INVALID_PARAMETER;
	}

	if (TTS_STATE_CREATED != client->current_state) {
		SLOG(LOG_ERROR, TAG_TTSC, "[ERROR] Unset error cb : Current state is not 'Created'.");
		return TTS_ERROR_INVALID_STATE;
	}

	client->error_cb = NULL;
	client->error_user_data = NULL;

	SLOG(LOG_DEBUG, TAG_TTSC, "[SUCCESS] Unset error cb");

	return 0;
}

int tts_set_default_voice_changed_cb(tts_h tts, tts_default_voice_changed_cb callback, void* user_data)
{
	if (0 != __tts_get_feature_enabled()) {
		return TTS_ERROR_NOT_SUPPORTED;
	}

	if (NULL == tts || NULL == callback) {
		SLOG(LOG_ERROR, TAG_TTSC, "[ERROR] Set default voice changed cb : Input parameter is null");
		return TTS_ERROR_INVALID_PARAMETER;
	}

	tts_client_s* client = tts_client_get(tts);

	if (NULL == client) {
		SLOG(LOG_ERROR, TAG_TTSC, "[ERROR] Set default voice changed cb : A handle is not valid");
		return TTS_ERROR_INVALID_PARAMETER;
	}

	if (TTS_STATE_CREATED != client->current_state) {
		SLOG(LOG_ERROR, TAG_TTSC, "[ERROR] Set default voice changed cb : Current state is not 'Created'.");
		return TTS_ERROR_INVALID_STATE;
	}

	client->default_voice_changed_cb = callback;
	client->default_voice_changed_user_data = user_data;

	SLOG(LOG_DEBUG, TAG_TTSC, "[SUCCESS] Set default voice changed cb");

	return 0;
}

int tts_unset_default_voice_changed_cb(tts_h tts)
{
	if (0 != __tts_get_feature_enabled()) {
		return TTS_ERROR_NOT_SUPPORTED;
	}

	if (NULL == tts) {
		SLOG(LOG_ERROR, TAG_TTSC, "[ERROR] Unset default voice changed cb : Input parameter is null");
		return TTS_ERROR_INVALID_PARAMETER;
	}

	tts_client_s* client = tts_client_get(tts);

	if (NULL == client) {
		SLOG(LOG_ERROR, TAG_TTSC, "[ERROR] Unset default voice changed cb : A handle is not valid");
		return TTS_ERROR_INVALID_PARAMETER;
	}

	if (TTS_STATE_CREATED != client->current_state) {
		SLOG(LOG_ERROR, TAG_TTSC, "[ERROR] Unset default voice changed cb : Current state is not 'Created'.");
		return TTS_ERROR_INVALID_STATE;
	}

	client->default_voice_changed_cb = NULL;
	client->default_voice_changed_user_data = NULL;

	SLOG(LOG_DEBUG, TAG_TTSC, "[SUCCESS] Unset default voice changed cb");

	return 0;
}

int tts_set_engine_changed_cb(tts_h tts, tts_engine_changed_cb callback, void* user_data)
{
	if (0 != __tts_get_feature_enabled()) {
		return TTS_ERROR_NOT_SUPPORTED;
	}

	if (NULL == tts || NULL == callback) {
		SLOG(LOG_ERROR, TAG_TTSC, "[ERROR] Set engine changed cb : Input parameter is null");
		return TTS_ERROR_INVALID_PARAMETER;
	}

	tts_client_s* client = tts_client_get(tts);

	if (NULL == client) {
		SLOG(LOG_ERROR, TAG_TTSC, "[ERROR] Set engine changed cb : A handle is not valid");
		return TTS_ERROR_INVALID_PARAMETER;
	}

	if (TTS_STATE_CREATED != client->current_state) {
		SLOG(LOG_ERROR, TAG_TTSC, "[ERROR] Set engine changed cb : Current state is not 'Created'.");
		return TTS_ERROR_INVALID_STATE;
	}

	client->engine_changed_cb = callback;
	client->engine_changed_user_data = user_data;

	SLOG(LOG_DEBUG, TAG_TTSC, "[SUCCESS] Set engine changed cb");

	return 0;
}

int tts_unset_engine_changed_cb(tts_h tts)
{
	if (0 != __tts_get_feature_enabled()) {
		return TTS_ERROR_NOT_SUPPORTED;
	}

	if (NULL == tts) {
		SLOG(LOG_ERROR, TAG_TTSC, "[ERROR] Unset engine changed cb : Input parameter is null");
		return TTS_ERROR_INVALID_PARAMETER;
	}

	tts_client_s* client = tts_client_get(tts);

	if (NULL == client) {
		SLOG(LOG_ERROR, TAG_TTSC, "[ERROR] Unset engine changed cb : A handle is not valid");
		return TTS_ERROR_INVALID_PARAMETER;
	}

	if (TTS_STATE_CREATED != client->current_state) {
		SLOG(LOG_ERROR, TAG_TTSC, "[ERROR] Unset engine changed cb : Current state is not 'Created'.");
		return TTS_ERROR_INVALID_STATE;
	}

	client->engine_changed_cb = NULL;
	client->engine_changed_user_data = NULL;

	SLOG(LOG_DEBUG, TAG_TTSC, "[SUCCESS] Unset engine changed cb");

	return 0;
}

int tts_add_pcm(tts_h tts, int event, const void* data, unsigned int data_size, int audio_type, int rate)
{
	if (0 != __tts_get_feature_enabled()) {
		return TTS_ERROR_NOT_SUPPORTED;
	}

	SLOG(LOG_INFO, TAG_TTSC, "@@@ Add pcm tts");

	if (NULL == tts) {
		SLOG(LOG_ERROR, TAG_TTSC, "[ERROR] Input handle is null.");
		SLOG(LOG_DEBUG, TAG_TTSC, "@@@");
		return TTS_ERROR_INVALID_PARAMETER;
	}

	tts_client_s* client = tts_client_get(tts);

	if (NULL == client) {
		SLOG(LOG_ERROR, TAG_TTSC, "[ERROR] A handle is not valid.");
		SLOG(LOG_DEBUG, TAG_TTSC, "@@@");
		return TTS_ERROR_INVALID_PARAMETER;
	}

	if (TTS_STATE_CREATED == client->current_state) {
		SLOG(LOG_ERROR, TAG_TTSC, "[ERROR] The current state is invalid.");
		return TTS_ERROR_INVALID_STATE;
	}

	if (false == g_screen_reader && TTS_MODE_SCREEN_READER == client->mode) {
		SLOG(LOG_WARN, TAG_TTSC, "[WARNING] Screen reader option is NOT available. Ignore this request");
		return TTS_ERROR_INVALID_STATE;
	}

	int ret = -1;
	int count = 0;
	bool is_prepared = false;
	while (0 != ret) {
		ret = tts_dbus_request_add_pcm(client->uid, event, data, data_size, audio_type, rate);
		if (0 != ret) {
			if (TTS_ERROR_INVALID_PARAMETER == ret && false == is_prepared) {
				client->current_state = TTS_STATE_CREATED;
				if (0 == tts_prepare_sync(tts)) {
					is_prepared = true;
					SLOG(LOG_INFO, TAG_TTSC, "[INFO] Success tts_prepare_sync");
				}
			} else if (TTS_ERROR_TIMED_OUT != ret) {
				SLOG(LOG_ERROR, TAG_TTSC, "[ERROR] result : %s", __tts_get_error_code(ret));
				return ret;
			} else {
				SLOG(LOG_WARN, TAG_TTSC, "[WARNING] retry add pcm : %s", __tts_get_error_code(ret));
				usleep(10000);
				count++;
				if (TTS_RETRY_COUNT == count) {
					SLOG(LOG_ERROR, TAG_TTSC, "[ERROR] Fail to request");
					return ret;
				}
			}
		}
	}

	SLOG(LOG_DEBUG, TAG_TTSC, "@@@");

	return TTS_ERROR_NONE;
}

int tts_play_pcm(tts_h tts)
{
	if (0 != __tts_get_feature_enabled()) {
		return TTS_ERROR_NOT_SUPPORTED;
	}

	SLOG(LOG_INFO, TAG_TTSC, "@@@ Play pcm tts");

	if (NULL == tts) {
		SLOG(LOG_ERROR, TAG_TTSC, "[ERROR] Input handle is null.");
		SLOG(LOG_DEBUG, TAG_TTSC, "@@@");
		return TTS_ERROR_INVALID_PARAMETER;
	}

	tts_client_s* client = tts_client_get(tts);

	if (NULL == client) {
		SLOG(LOG_ERROR, TAG_TTSC, "[ERROR] A handle is not valid.");
		SLOG(LOG_DEBUG, TAG_TTSC, "@@@");
		return TTS_ERROR_INVALID_PARAMETER;
	}

	if (TTS_STATE_PLAYING == client->current_state || TTS_STATE_CREATED == client->current_state) {
		SLOG(LOG_ERROR, TAG_TTSC, "[ERROR] The current state is invalid.");
		return TTS_ERROR_INVALID_STATE;
	}

	if (false == g_screen_reader && TTS_MODE_SCREEN_READER == client->mode) {
		SLOG(LOG_WARN, TAG_TTSC, "[WARNING] Screen reader option is NOT available. Ignore this request");
		return TTS_ERROR_INVALID_STATE;
	}

	int ret = -1;
	int count = 0;
	bool is_prepared = false;
	while (0 != ret) {
		ret = tts_dbus_request_play_pcm(client->uid);
		if (0 != ret) {
			if (TTS_ERROR_INVALID_PARAMETER == ret && false == is_prepared) {
				client->current_state = TTS_STATE_CREATED;
				if (0 == tts_prepare_sync(tts)) {
					is_prepared = true;
					SLOG(LOG_INFO, TAG_TTSC, "[INFO] Success tts_prepare_sync");
				}
			} else if (TTS_ERROR_TIMED_OUT != ret) {
				SLOG(LOG_ERROR, TAG_TTSC, "[ERROR] result : %s", __tts_get_error_code(ret));
				return ret;
			} else {
				SLOG(LOG_WARN, TAG_TTSC, "[WARNING] retry play pcm : %s", __tts_get_error_code(ret));
				usleep(10000);
				count++;
				if (TTS_RETRY_COUNT == count) {
					SLOG(LOG_ERROR, TAG_TTSC, "[ERROR] Fail to request");
					return ret;
				}
			}
		}
	}

	client->before_state = client->current_state;
	client->current_state = TTS_STATE_PLAYING;

	if (NULL != client->state_changed_cb) {
		tts_client_use_callback(client);
		client->state_changed_cb(client->tts, client->before_state, client->current_state, client->state_changed_user_data);
		tts_client_not_use_callback(client);
		SLOG(LOG_DEBUG, TAG_TTSC, "State changed callback is called");
	}

	SLOG(LOG_DEBUG, TAG_TTSC, "@@@");

	return TTS_ERROR_NONE;
}

int tts_stop_pcm(tts_h tts)
{
	if (0 != __tts_get_feature_enabled()) {
		return TTS_ERROR_NOT_SUPPORTED;
	}

	SLOG(LOG_INFO, TAG_TTSC, "@@@ Stop pcm tts");

	if (NULL == tts) {
		SLOG(LOG_ERROR, TAG_TTSC, "[ERROR] Input handle is null.");
		SLOG(LOG_DEBUG, TAG_TTSC, "@@@");
		return TTS_ERROR_INVALID_PARAMETER;
	}

	tts_client_s* client = tts_client_get(tts);

	if (NULL == client) {
		SLOG(LOG_ERROR, TAG_TTSC, "[ERROR] A handle is not valid.");
		SLOG(LOG_DEBUG, TAG_TTSC, "@@@");
		return TTS_ERROR_INVALID_PARAMETER;
	}

	if (TTS_STATE_CREATED == client->current_state) {
		SLOG(LOG_ERROR, TAG_TTSC, "[ERROR] The current state is invalid.");
		return TTS_ERROR_INVALID_STATE;
	}

	if (false == g_screen_reader && TTS_MODE_SCREEN_READER == client->mode) {
		SLOG(LOG_WARN, TAG_TTSC, "[WARNING] Screen reader option is NOT available. Ignore this request");
		return TTS_ERROR_INVALID_STATE;
	}

	int ret = -1;
	int count = 0;
	bool is_prepared = false;
	while (0 != ret) {
		ret = tts_dbus_request_stop_pcm(client->uid);
		if (0 != ret) {
			if (TTS_ERROR_INVALID_PARAMETER == ret && false == is_prepared) {
				client->current_state = TTS_STATE_CREATED;
				if (0 == tts_prepare_sync(tts)) {
					is_prepared = true;
					SLOG(LOG_INFO, TAG_TTSC, "[INFO] Success tts_prepare_sync");
				}
			} else if (TTS_ERROR_TIMED_OUT != ret) {
				SLOG(LOG_ERROR, TAG_TTSC, "[ERROR] result : %s", __tts_get_error_code(ret));
				return ret;
			} else {
				SLOG(LOG_WARN, TAG_TTSC, "[WARNING] retry stop pcm : %s", __tts_get_error_code(ret));
				usleep(10000);
				count++;
				if (TTS_RETRY_COUNT == count) {
					SLOG(LOG_ERROR, TAG_TTSC, "[ERROR] Fail to request");
					return ret;
				}
			}
		}
	}

	client->before_state = client->current_state;
	client->current_state = TTS_STATE_READY;

	if (NULL != client->state_changed_cb) {
		tts_client_use_callback(client);
		client->state_changed_cb(client->tts, client->before_state, client->current_state, client->state_changed_user_data);
		tts_client_not_use_callback(client);
		SLOG(LOG_DEBUG, TAG_TTSC, "State changed callback is called");
	}

	SLOG(LOG_DEBUG, TAG_TTSC, "@@@");

	return TTS_ERROR_NONE;
}

int tts_repeat(tts_h tts, char** text_repeat, int* utt_id)
{
	if (0 != __tts_get_feature_enabled()) {
		return TTS_ERROR_NOT_SUPPORTED;
	}

	SLOG(LOG_INFO, TAG_TTSC, "@@@ Repeat TTS");

	if (NULL == tts) {
		SLOG(LOG_ERROR, TAG_TTSC, "[ERROR] Input handle is null.");
		SLOG(LOG_DEBUG, TAG_TTSC, "@@@");
		return TTS_ERROR_INVALID_PARAMETER;
	}

	if (NULL == text_repeat || NULL == utt_id) {
		SLOG(LOG_ERROR, TAG_TTSC, "[ERROR] Input parameter is null.");
		SLOG(LOG_DEBUG, TAG_TTSC, "@@@");
		return TTS_ERROR_INVALID_PARAMETER;
	}

	*text_repeat = NULL;
	*utt_id = -1;

	tts_client_s* client = tts_client_get(tts);

	if (NULL == client) {
		SLOG(LOG_ERROR, TAG_TTSC, "[ERROR] A handle is not valid.");
		SLOG(LOG_DEBUG, TAG_TTSC, "@@@");
		return TTS_ERROR_INVALID_PARAMETER;
	}

	if (TTS_STATE_READY != client->current_state) {
		SLOG(LOG_ERROR, TAG_TTSC, "[ERROR] The current state is invalid. (%d)", client->current_state);
		return TTS_ERROR_INVALID_STATE;
	}

	/* Clear the legacy and Add texts to be played repeatedly */
	int ret = -1;
	ret = tts_stop(tts);
	if (TTS_ERROR_NONE != ret) {
		SLOG(LOG_ERROR, TAG_TTSC, "[ERROR] Fail to clear the legacy");
		return ret;
	}

	if (NULL != client->text_repeat) {
		char* tmp_text = strdup(client->text_repeat);
		char* tmp_lang = NULL;
		if (NULL != g_language) {
			tmp_lang = strdup(g_language);
		}
		ret = tts_add_text(tts, tmp_text, tmp_lang, g_voice_type, g_speed, utt_id);
		if (TTS_ERROR_NONE != ret) {
			SLOG(LOG_ERROR, TAG_TTSC, "[ERROR] Fail to add texts for repetition.");
			if (NULL != tmp_text) {
				free(tmp_text);
				tmp_text = NULL;
			}
			if (NULL != tmp_lang) {
				free(tmp_lang);
				tmp_lang = NULL;
			}
			return ret;
		}
		*text_repeat = strdup(client->text_repeat);
		SLOG(LOG_DEBUG, TAG_TTSC, "[DEBUG] text to repeat(%s), utt_id(%d)", *text_repeat, *utt_id);
		if (NULL != tmp_text) {
			free(tmp_text);
			tmp_text = NULL;
		}
		if (NULL != tmp_lang) {
			free(tmp_lang);
			tmp_lang = NULL;
		}
	} else {
		SLOG(LOG_ERROR, TAG_TTSC, "[ERROR] There is no previous added texts. Please add texts");
		return TTS_ERROR_OPERATION_FAILED;
	}

	/* Play added texts */
	ret = tts_play(tts);
	if (TTS_ERROR_NONE != ret) {
		SLOG(LOG_ERROR, TAG_TTSC, "[ERROR] Fail to play texts for repetition.");
		if (NULL != *text_repeat) {
			free(*text_repeat);
			*text_repeat = NULL;
		}
		*utt_id = -1;
		return ret;
	}

	return TTS_ERROR_NONE;
}
